import os

import matplotlib.pyplot as pl
import numpy as np
import pandas as pd
import warnings

from aethos.config import IMAGE_DIR

warnings.simplefilter("ignore", UserWarning)


class Shap(object):
    def __init__(
        self, model, model_name, x_train, x_test, y_test, learner: str, shap_values
    ):

        import lightgbm as lgb
        import shap

        self.model = model
        self.model_name = model_name
        self.x_train = x_train
        self.x_test = x_test
        self.y_test = y_test

        if learner == "linear":
            self.explainer = shap.LinearExplainer(
                self.model, self.x_train, feature_dependence="independent"
            )
        elif learner == "tree":
            self.explainer = shap.TreeExplainer(self.model)
        elif learner == "kernel":
            if hasattr(self.model, "predict_proba"):
                func = self.model.predict_proba
            else:
                func = self.model.predict

            self.explainer = shap.KernelExplainer(func, self.x_train)
        else:
            raise ValueError(f"Learner: {learner} is not supported yet.")

        if shap_values is None:
            self.expected_value = self.explainer.expected_value
            self.shap_values = np.array(self.explainer.shap_values(self.x_test)).astype(
                float
            )
        else:
            self.expected_value = shap_values[0, -1]
            self.shap_values = shap_values[:, :-1]

        if isinstance(self.model, lgb.sklearn.LGBMClassifier) and isinstance(
            self.expected_value, np.float
        ):
            self.shap_values = self.shap_values[1]

        # Calculate misclassified values
        self.misclassified_values = self._calculate_misclassified()

        # As per SHAP guidelines, test data needs to be dense for plotting functions
        self.x_test_array = self.x_test.values

    def summary_plot(self, output_file="", **summaryplot_kwargs):
        """
        Plots a SHAP summary plot.

        Parameters
        ----------
        output_file: str
            Output file name including extension (.png, .jpg, etc.) to save image as.
        """

        import shap

        shap.summary_plot(
            self.shap_values,
            self.x_test_array,
            feature_names=self.x_train.columns,
            show=False,
            **summaryplot_kwargs,
        )

        if output_file:  # pragma: no cover
            pl.savefig(os.path.join(IMAGE_DIR, self.model_name, output_file))

    def decision_plot(
        self, num_samples=0.25, sample_no=None, output_file="", **decisionplot_kwargs
    ):
        """
        Plots a SHAP decision plot.
        
        Parameters
        ----------
        num_samples : int, float, or 'all', optional
            Number of samples to display, if less than 1 it will treat it as a percentage, 'all' will include all samples
            , by default 0.25

        sample_no : int, optional
            Sample number to isolate and analyze, if provided it overrides num_samples, by default None

        Returns
        -------
        DecisionPlotResult 
            If return_objects=True (the default). Returns None otherwise.
        """

        import shap

        return_objects = decisionplot_kwargs.pop("return_objects", True)
        highlight = decisionplot_kwargs.pop("highlight", None)

        if sample_no is not None:
            if sample_no < 1 or not isinstance(sample_no, int):
                raise ValueError("Sample number must be greater than 1.")

            samples = slice(sample_no - 1, sample_no)
        else:
            if num_samples == "all":
                samples = slice(0, len(self.x_test_array))
            elif num_samples <= 0:
                raise ValueError(
                    "Number of samples must be greater than 0. If it is less than 1, it will be treated as a percentage."
                )
            elif num_samples > 0 and num_samples < 1:
                samples = slice(0, int(num_samples * len(self.x_test_array)))
            else:
                samples = slice(0, num_samples)

        if highlight is not None:
            highlight = highlight[samples]

        s = shap.decision_plot(
            self.expected_value,
            self.shap_values[samples],
            self.x_train.columns,
            return_objects=return_objects,
            highlight=highlight,
            show=False,
            **decisionplot_kwargs,
        )

        if output_file:  # pragma: no cover
            pl.savefig(os.path.join(IMAGE_DIR, self.model_name, output_file))

        return s

    def force_plot(self, sample_no=None, output_file="", **forceplot_kwargs):
        """
        Plots a SHAP force plot.
        """

        import shap

        shap_values = forceplot_kwargs.pop("shap_values", self.shap_values)

        if sample_no is not None:
            if sample_no < 1 or not isinstance(sample_no, int):
                raise ValueError("Sample number must be greater than 1.")

            samples = slice(sample_no - 1, sample_no)
        else:
            samples = slice(0, len(shap_values))

        s = shap.force_plot(
            self.expected_value,
            shap_values[samples],
            self.x_train.columns,
            **forceplot_kwargs,
        )

        if output_file:  # pragma: no cover
            pl.savefig(os.path.join(IMAGE_DIR, self.model_name, output_file))

        return s

    def dependence_plot(
        self, feature, interaction=None, output_file="", **dependenceplot_kwargs
    ):
        """
        Plots a SHAP dependence plot.
        """

        import shap

        interaction = dependenceplot_kwargs.pop("interaction_index", interaction)

        shap.dependence_plot(
            feature,
            self.shap_values,
            self.x_test,
            interaction_index=interaction,
            show=False,
            **dependenceplot_kwargs,
        )

        if output_file:  # pragma: no cover
            pl.savefig(os.path.join(IMAGE_DIR, self.model_name, output_file))

    def _calculate_misclassified(self) -> list:
        """
        Calculates misclassified points.
        
        Returns
        -------
        list
            List specifying which values were misclassified
        """

        if len(self.shap_values.shape) > 2:
            y_pred = [
                x.sum(1) + y > 0 for x, y in zip(self.shap_values, self.expected_value)
            ]
            misclassified = [x != self.y_test for x in y_pred]
        else:
            y_pred = (self.shap_values.sum(1) + self.expected_value) > 0
            misclassified = y_pred != self.y_test

        return misclassified


class MSFTInterpret(object):
    def __init__(self, model, x_train, x_test, y_test, problem):

        self.model = model
        self.x_train = x_train.apply(pd.to_numeric)
        self.x_test = x_test.apply(pd.to_numeric)
        self.y_test = y_test
        self.problem = problem
        self.trained_blackbox_explainers = {}

    def blackbox_show_performance(self, method, predictions="default", show=True):
        """
        Plots an interpretable display of your model based off a performance metric.

        Can either be 'ROC' or 'PR' for precision, recall for classification problems.

        Can be 'regperf' for regression problems.
        
        Parameters
        ----------
        method : str
            Performance metric, either 'roc' or 'PR'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional
            False to not display the plot, by default True
        
        Returns
        -------
        Interpret
            Interpretable dashboard of your model
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        if predictions == "probability":
            predict_fn = self.model.predict_proba
        else:
            predict_fn = self.model.predict

        if self.problem in INTERPRET_EXPLAINERS["problem"]:
            if method.lower() in INTERPRET_EXPLAINERS["problem"][self.problem]:
                blackbox_perf = INTERPRET_EXPLAINERS["problem"][self.problem][
                    method.lower()
                ](predict_fn).explain_perf(
                    self.x_test, self.y_test, name=method.upper()
                )
        else:
            raise ValueError(
                "Supported blackbox explainers are only {} for classification problems and {} for regression problems".format(
                    ",".join(INTERPRET_EXPLAINERS["problem"]["classification"].keys()),
                    ",".join(INTERPRET_EXPLAINERS["problem"]["regression"].keys()),
                )
            )

        if show:
            interpret.show(blackbox_perf)

        self.trained_blackbox_explainers[method.lower()] = blackbox_perf

        return blackbox_perf

    def blackbox_local_explanation(
        self,
        num_samples=0.5,
        sample_no=None,
        method="lime",
        predictions="default",
        show=True,
        **kwargs,
    ):
        """
        Plots an interpretable display that explains individual predictions of your model.

        Supported explainers are either 'lime' or 'shap'.
        
        Parameters
        ----------
        num_samples : int, float, or 'all', optional
            Number of samples to display, if less than 1 it will treat it as a percentage, 'all' will include all samples
            , by default 0.25

        sample_no : int, optional
            Sample number to isolate and analyze, if provided it overrides num_samples, by default None

        method : str, optional
            Explainer type, can either be 'lime' or 'shap', by default 'lime'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional
            False to not display the plot, by default True
        
        Returns
        -------
        Interpret
            Interpretable dashboard of your model
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        if predictions == "probability":
            predict_fn = self.model.predict_proba
        else:
            predict_fn = self.model.predict

        # Determine method
        if method.lower() in INTERPRET_EXPLAINERS["local"]:
            if method.lower() == "lime":
                data = self.x_train
            elif method.lower() == "shap":
                data = np.median(self.x_train, axis=0).reshape(1, -1)
            else:
                raise ValueError

            explainer = INTERPRET_EXPLAINERS["local"][method.lower()](
                predict_fn=predict_fn,
                data=data,
                feature_names=self.x_train.columns.tolist(),
                **kwargs,
            )

        else:
            raise ValueError(
                'Supported blackbox local explainers are only "lime" and "shap".'
            )

        if sample_no is not None:
            if sample_no < 1 or not isinstance(sample_no, int):
                raise ValueError("Sample number must be greater than 1.")

            samples = slice(sample_no - 1, sample_no)
        else:
            if num_samples == "all":
                samples = slice(0, len(self.x_test))
            elif num_samples <= 0:
                raise ValueError(
                    "Number of samples must be greater than 0. If it is less than 1, it will be treated as a percentage."
                )
            elif num_samples > 0 and num_samples < 1:
                samples = slice(0, int(num_samples * len(self.x_test)))
            else:
                samples = slice(0, num_samples)

        explainer_local = explainer.explain_local(
            self.x_test[samples], self.y_test[samples], name=method.upper()
        )

        self.trained_blackbox_explainers[method.lower()] = explainer_local

        if show:
            interpret.show(explainer_local)

        return explainer_local

    def blackbox_global_explanation(
        self, method="morris", predictions="default", show=True, **kwargs
    ):
        """
        Provides an interpretable summary of your models behaviour based off an explainer.

        Can either be 'morris' or 'dependence' for Partial Dependence.
        
        Parameters
        ----------
        method : str, optional
            Explainer type, can either be 'morris' or 'dependence', by default 'morris'

        predictions : str, optional
            Prediction type, can either be 'default' (.predict) or 'probability' if the model can predict probabilities, by default 'default'

        show : bool, optional
            False to not display the plot, by default True
        
        Returns
        -------
        Interpret
            Interpretable dashboard of your model
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        if predictions == "probability":
            predict_fn = self.model.predict_proba
        else:
            predict_fn = self.model.predict

        if method.lower() in INTERPRET_EXPLAINERS["global"]:
            sensitivity = INTERPRET_EXPLAINERS["global"][method.lower()](
                predict_fn=predict_fn, data=self.x_train, **kwargs
            )

        else:
            raise ValueError(
                'Supported blackbox global explainers are only "morris" and partial "dependence".'
            )

        sensitivity_global = sensitivity.explain_global(name=method.upper())

        self.trained_blackbox_explainers[method.lower()] = sensitivity_global

        if show:
            interpret.show(sensitivity_global)

        return sensitivity_global

    def create_dashboard(self):  # pragma: no cover
        """
        Displays an interpretable dashboard of already created interpretable plots.
        
        If a plot hasn't been interpreted yet it is created using default parameters for the dashboard.
        """

        import interpret
        from aethos.modelling.constants import INTERPRET_EXPLAINERS

        dashboard_plots = []

        for explainer_type in INTERPRET_EXPLAINERS:

            if explainer_type == "problem":
                temp_explainer_type = INTERPRET_EXPLAINERS[explainer_type][self.problem]
            else:
                temp_explainer_type = INTERPRET_EXPLAINERS[explainer_type]

            for explainer in temp_explainer_type:
                if explainer in self.trained_blackbox_explainers:
                    dashboard_plots.append(self.trained_blackbox_explainers[explainer])
                else:
                    if explainer_type == "problem":
                        dashboard_plots.append(
                            self.blackbox_show_performance(explainer, show=False)
                        )
                    elif explainer_type == "local":
                        dashboard_plots.append(
                            self.blackbox_local_explanation(
                                method=explainer, show=False
                            )
                        )
                    else:
                        dashboard_plots.append(
                            self.blackbox_global_explanation(
                                method=explainer, show=False
                            )
                        )

        interpret.show(dashboard_plots)



from aethos.visualizations import visualize as viz
from typing import Union
import numpy as np


class Visualizations(object):

    @property
    def plot_colors(self):  # pragma: no cover
        """
        Displays all plot colour names
        """

        from IPython.display import IFrame

        IFrame(
            "https://python-graph-gallery.com/wp-content/uploads/100_Color_names_python.png"
        )

    @property
    def plot_colorpalettes(self):  # pragma: no cover
        """
        Displays color palette configuration guide.
        """

        from IPython.display import IFrame

        IFrame("https://seaborn.pydata.org/tutorial/color_palettes.html")

    @property
    def train_data(self): # pragma: no cover
        if hasattr(self, '_train_result_data'):
            return self._train_result_data
        else:
            return self.x_train

    @property
    def test_data(self): # pragma: no cover
        if hasattr(self, '_test_result_data'):
            return self._test_result_data
        else:
            return self.x_test

    def raincloud(self, x=None, y=None, output_file="", **params):
        """
        Combines the box plot, scatter plot and split violin plot into one data visualization.
        This is used to offer eyeballed statistical inference, assessment of data distributions (useful to check assumptions),
        and the raw data itself showing outliers and underlying patterns.

        A raincloud is made of:
        1) "Cloud", kernel desity estimate, the half of a violinplot.
        2) "Rain", a stripplot below the cloud
        3) "Umberella", a boxplot
        4) "Thunder", a pointplot connecting the mean of the different categories (if `pointplot` is `True`)

        Useful parameter documentation
        ------------------------------
        https://seaborn.pydata.org/generated/seaborn.boxplot.html

        https://seaborn.pydata.org/generated/seaborn.violinplot.html

        https://seaborn.pydata.org/generated/seaborn.stripplot.html

        Parameters
        ----------
        x : str
            X axis data, reference by column name, any data

        y : str
            Y axis data, reference by column name, measurable data (numeric)
            by default target_field

        hue : Iterable, np.array, or dataframe column name if 'data' is specified
            Second categorical data. Use it to obtain different clouds and rainpoints

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        orient : str                  
            vertical if "v" (default), horizontal if "h"

        width_viol : float            
            width of the cloud

        width_box : float             
            width of the boxplot

        palette : list or dict        
            Colours to use for the different levels of categorical variables

        bw : str or float
            Either the name of a reference rule or the scale factor to use when computing the kernel bandwidth,
            by default "scott"

        linewidth : float             
            width of the lines

        cut : float
            Distance, in units of bandwidth size, to extend the density past the extreme datapoints.
            Set to 0 to limit the violin range within the range of the observed data,
            by default 2

        scale : str
            The method used to scale the width of each violin.
            If area, each violin will have the same area.
            If count, the width of the violins will be scaled by the number of observations in that bin.
            If width, each violin will have the same width.
            By default "area"

        jitter : float, True/1
            Amount of jitter (only along the categorical axis) to apply.
            This can be useful when you have many points and they overlap,
            so that it is easier to see the distribution. You can specify the amount of jitter (half the width of the uniform random variable support),
            or just use True for a good default.

        move : float                  
            adjust rain position to the x-axis (default value 0.)

        offset : float                
            adjust cloud position to the x-axis

        color : matplotlib color
            Color for all of the elements, or seed for a gradient palette.

        ax : matplotlib axes
            Axes object to draw the plot onto, otherwise uses the current Axes.

        figsize : (int, int)    
            size of the visualization, ex (12, 5)

        pointplot : bool   
            line that connects the means of all categories, by default False

        dodge : bool 
            When hue nesting is used, whether elements should be shifted along the categorical axis.

        Source: https://micahallen.org/2018/03/15/introducing-raincloud-plots/
        
        Examples
        --------
        >>> data.raincloud('col1') # Will plot col1 values on the x axis and your target variable values on the y axis
        >>> data.raincloud('col1', 'col2') # Will plot col1 on the x and col2 on the y axis
        >>> data.raincloud('col1', 'col2', output_file='raincloud.png')
        """

        if y is None:
            y = self.target_field

        fig = viz.raincloud(y, x, self.train_data, output_file=output_file, **params)

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def barplot(
        self,
        x: str,
        y: Union[str, list],
        method=None,
        orient="v",
        barmode='relative',
        title='',
        yaxis_params=None,
        xaxis_params=None,
        output_file="",
        **barplot_kwargs,
    ):
        """
        Plots a bar plot for the given columns provided using Plotly.

        If `groupby` is provided, method must be provided for example you may want to plot Age against survival rate,
        so you would want to `groupby` Age and then find the `mean` as the method.

        For a list of group by methods please checkout the following pandas link:
        
            https://pandas.pydata.org/pandas-docs/stable/reference/groupby.html#computations-descriptive-stats

        For a list of possible arguments for the bar plot please checkout the following links:

            https://plot.ly/python/reference/#bar

        Parameters
        ----------
        x : str
            Column name for the x axis.

        y : str or list
            Column(s) you would like to see plotted against the x_col

        method : str
            Method to aggregate groupy data
            Examples: min, max, mean, etc., optional
            by default None

        orient : str, optional
            Orientation of graph, 'h' for horizontal
            'v' for vertical, by default 'v',

        barmode : str {'relative', 'overlay', 'group', 'stack'}
            Relative is a normal barplot
            Overlay barplot shows positive values above 0 and negative values below 0
            Group are the bars beside each other.
            Stack groups the bars on top of each other
            by default 'relative'

        title : str
            Title of the plot, by default ''

        yaxis_params : dict
            Parameters for the y axis, by default None

        xaxis_params : dict
            Parameters for the x axis, by default None

        output_file : str, optional
            Output html file name for image

        Examples
        --------
        >>> data.barplot(x='x', y='y')
        >>> data.barplot(x='x', y=['y', 'z'], method='mean')
        >>> data.barplot(x='x', y=['y', 'z'], method='max', orient='h')
        """

        fig = viz.barplot(
            x,
            y,
            self.train_data,
            method=method,
            orient=orient,
            barmode=barmode,
            output_file=output_file,
            **barplot_kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def scatterplot(
        self,
        x=None,
        y=None,
        z=None,
        category=None,
        title="Scatter Plot",
        output_file="",
        **scatterplot_kwargs,
    ):
        """
        Plots a scatterplot for the given x and y columns provided using Plotly Express.

        For a list of possible scatterplot_kwargs for 2 dimensional data please check out the following links:

            https://plot.ly/python-api-reference/generated/plotly.express.scatter.html

        For more information on key word arguments for 3d data, please check them out here:

            https://www.plotly.express/plotly_express/#plotly_express.scatter_3d
        
        Parameters
        ----------
        x : str
            X column name

        y : str
            Y column name

        z : str
            Z column name, 

        category : str, optional
            Category to group your data, by default None

        title : str, optional
            Title of the plot, by default 'Scatter Plot'

        output_file : str, optional
            Output html file name for image

        **scatterplot_kwargs : optional
            See above links for list of possible scatterplot options.

        Examples
        --------
        >>> data.scatterplot(x='x', y='y') #2d
        >>> data.scatterplot(x='x', y='y', z='z') #3d
        >>> data.scatterplot(x='x', y='y', z='z', output_file='scatt')
        """

        fig = viz.scatterplot(
            x,
            y,
            z=z,
            data=self.train_data,
            title=title,
            category=category,
            output_file=output_file,
            **scatterplot_kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def lineplot(
        self, x: str, y: str, z=None, category=None, title="Line Plot", output_file="", **lineplot_kwargs
    ):
        """
        Plots a lineplot for the given x and y columns provided using Plotly Express.

        For a list of possible lineplot_kwargs please check out the following links:

        For 2d:

            https://plot.ly/python-api-reference/generated/plotly.express.line.html#plotly.express.line

        For 3d:

            https://plot.ly/python-api-reference/generated/plotly.express.line_3d.html#plotly.express.line_3d
        
        Parameters
        ----------
        x : str
            X column name

        y : str
            Column name to plot on the y axis.

        z: str
            Column name to plot on the z axis.

        title : str, optional
            Title of the plot, by default 'Line Plot'

        category : str
            Category column to draw multiple line plots of

        output_file : str, optional
            Output html file name for image

        text : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like appear in the figure as text labels.

        facet_row : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to assign marks to facetted subplots in the vertical direction.

        facet_col : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to assign marks to facetted subplots in the horizontal direction.

        error_x : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to size x-axis error bars.
            If error_x_minus is None, error bars will be symmetrical, otherwise error_x is used for the positive direction only.

        error_x_minus : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to size x-axis error bars in the negative direction.
            Ignored if error_x is None.

        error_y : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to size y-axis error bars.
            If error_y_minus is None, error bars will be symmetrical, otherwise error_y is used for the positive direction only.

        error_y_minus : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to size y-axis error bars in the negative direction.
            Ignored if error_y is None.

        animation_frame : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to assign marks to animation frames.

        animation_group : str or int or Series or array-like
            Either a name of a column in data_frame, or a pandas Series or array_like object.
            Values from this column or array_like are used to provide object-constancy across animation frames: rows with matching `animation_group`s will be treated as if they describe the same object in each frame.

        labels : dict with str keys and str values (default {})
            By default, column names are used in the figure for axis titles, legend entries and hovers. 
            his parameter allows this to be overridden. The keys of this dict should correspond to column names, and the values should correspond to the desired label to be displayed.

        color_discrete_sequence : list of str
            Strings should define valid CSS-colors. 
            When color is set and the values in the corresponding column are not numeric, values in that column are assigned colors by cycling through color_discrete_sequence in the order described in category_orders, unless the value of color is a key in color_discrete_map.
            Various useful color sequences are available in the plotly.express.colors submodules, specifically plotly.express.colors.qualitative.

        color_discrete_map : dict with str keys and str values (default {})
            String values should define valid CSS-colors Used to override color_discrete_sequence to assign a specific colors to marks corresponding with specific values.
            Keys in color_discrete_map should be values in the column denoted by color.

        Examples
        --------
        >>> data.line_plot(x='x', y='y')
        >>> data.line_plot(x='x', y='y', output_file='line')
        """

        fig = viz.lineplot(
            x, y, z, self.train_data, category=category, title=title, output_file=output_file, **lineplot_kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def correlation_matrix(
        self, data_labels=False, hide_mirror=False, output_file="", **kwargs
    ):
        """
        Plots a correlation matrix of all the numerical variables.

        For more information on possible kwargs please see: https://seaborn.pydata.org/generated/seaborn.heatmap.html
        
        Parameters
        ----------
        data_labels : bool, optional
            True to display the correlation values in the plot, by default False

        hide_mirror : bool, optional
            Whether to display the mirroring half of the correlation plot, by default False

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.correlation_matrix(data_labels=True)
        >>> data.correlation_matrix(data_labels=True, output_file='corr.png')
        """

        fig = viz.correlation_matrix(
            self.train_data,
            data_labels=data_labels,
            hide_mirror=hide_mirror,
            output_file=output_file,
            **kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def pairplot(
        self, cols=[], kind="scatter", diag_kind="auto", upper_kind=None, lower_kind=None, hue=None, output_file="", **kwargs
    ):
        """
        Plots pairplots of the variables from the training data.

        If hue is not provided and a target variable is set, the data will separated and highlighted by the classes in that column.

        For more info and kwargs on pair plots, please see: https://seaborn.pydata.org/generated/seaborn.pairplot.html
        
        Parameters
        ----------
        cols : list
            Columns to view pairplot of.

        kind : str {'scatter', 'reg'}, optional
            Type of plot for off-diag plots, by default 'scatter'

        diag_kind : str {'auto', 'hist', 'kde'}, optional
            Type of plot for diagonal, by default 'auto'

        upper_kind : str {'scatter', 'kde'}, optional
            Type of plot for upper triangle of pair plot, by default None

        lower_kind : str {'scatter', 'kde'}, optional
            Type of plot for lower triangle of pair plot, by default None

        hue : str, optional
            Column to colour points by, by default None

        {x, y}_vars : lists of variable names, optional
            Variables within data to use separately for the rows and columns of the figure; i.e. to make a non-square plot.

        palette : dict or seaborn color palette
            Set of colors for mapping the hue variable. If a dict, keys should be values in the hue variable.

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.pairplot(kind='kde')
        >>> data.pairplot(kind='kde', output_file='pair.png')
        """

        data = self.train_data if not cols else self.train_data[cols]

        fig = viz.pairplot(
            data,
            kind=kind,
            diag_kind=diag_kind,
            upper_kind=upper_kind,
            lower_kind=lower_kind,
            hue=hue,
            output_file=output_file,
            **kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def jointplot(self, x: str, y: str, kind="scatter", output_file="", **kwargs):
        """
        Plots joint plots of 2 different variables.

        Scatter ('scatter'): Scatter plot and histograms of x and y.

        Regression ('reg'): Scatter plot, with regression line and histograms with kernel density fits.

        Residuals ('resid'): Scatter plot of residuals and histograms of residuals.

        Kernel Density Estimates ('kde'): Density estimate plot and histograms.

        Hex ('hex'): Replaces scatterplot with joint histogram using hexagonal bins and histograms on the axes.

        For more info and kwargs for joint plots, see https://seaborn.pydata.org/generated/seaborn.jointplot.html
        
        Parameters
        ----------
        x : str
            X axis column

        y : str
            y axis column

        kind : { “scatter” | “reg” | “resid” | “kde” | “hex” }, optional
            Kind of plot to draw, by default 'scatter'

        color : matplotlib color, optional
            Color used for the plot elements.

        dropna : bool, optional
            If True, remove observations that are missing from x and y.

        {x, y}lim : two-tuples, optional
            Axis limits to set before plotting.

        {joint, marginal, annot}_kws : dicts, optional
            Additional keyword arguments for the plot components.

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.jointplot(x='x', y='y', kind='kde', color='crimson')
        >>> data.jointplot(x='x', y='y', kind='kde', color='crimson', output_file='pair.png')
        """

        fig = viz.jointplot(x=x, y=y, df=self.train_data, kind=kind, output_file=output_file, **kwargs)

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def histogram(self, *x, plot_test=False, output_file="", **kwargs):
        """
        Plots a histogram of the given column(s).

        If no columns are provided, histograms are plotted for all numeric columns

        For more histogram key word arguments, please see https://seaborn.pydata.org/generated/seaborn.distplot.html

        Parameters
        ----------
        x: str or str(s)
            Column(s) to plot histograms for.

        plot_test : bool, optional
            True to plot distribution of the test data for the same variable

        bins : argument for matplotlib hist(), or None, optional
            Specification of hist bins, or None to use Freedman-Diaconis rule.

        hist : bool, optional
            Whether to plot a (normed) histogram.

        kde : bool, optional
            Whether to plot a gaussian kernel density estimate.

        rug : bool, optional
            Whether to draw a rugplot on the support axis.

        fit : random variable object, optional
            An object with fit method, returning a tuple that can be passed to a pdf method a positional arguments following an grid of values to evaluate the pdf on.

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.histogram()
        >>> data.histogram('col1')
        >>> data.histogram('col1', 'col2', plot_test=True)
        >>> data.histogram('col1', kde=False)
        >>> data.histogram('col1', 'col2', hist=False)
        >>> data.histogram('col1', kde=False, fit=stat.normal)
        >>> data.histogram('col1', kde=False, output_file='hist.png')
        """

        x_test = self.test_data if plot_test else None
        columns = list(x) if x else list(self.train_data.select_dtypes(include=[np.number]).columns)

        viz.histogram(columns, x_train=self.train_data, x_test=x_test, output_file=output_file, **kwargs)

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

    def plot_dim_reduction(self, category: str, dim=2, algo='tsne', output_file="", **kwargs):
        """
        Reduce the dimensions of your data and then view similarly grouped data points (clusters)

        For 2d plotting options, see:
        
            https://plot.ly/python-api-reference/generated/plotly.express.scatter.html

        For 3d plotting options, see:

            https://www.plotly.express/plotly_express/#plotly_express.scatter_3d
        
        Parameters
        ----------
        category : str
            Column name of the labels/data points to highlight in the plot
            
        dim : int {2, 3}
            Dimensions of the plot to show, either 2d or 3d, by default 2

        algo : str {'tsne', 'lle', 'pca', 'tsvd'}, optional
            Algorithm to reduce the dimensions by, by default 'tsne'

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        kwargs: 
            See plotting options

        Examples
        --------
        >>> data.plot_dim_reduction('cluster_labels', dim=3)
        """

        fig = viz.viz_clusters(
            self.train_data,
            algo=algo,
            category=category,
            dim=dim,
            output_file=output_file,
            **kwargs,
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def boxplot(self, x=None, y=None, color=None, title="", output_file="", **kwargs):
        """
        Plots a box plot for the given x and y columns.

        For more info and kwargs for box plots, see https://plot.ly/python-api-reference/generated/plotly.express.box.html#plotly.express.box
        and https://plot.ly/python/box-plots/ 
        
        Parameters
        ----------
        x : str
            X axis column

        y : str
            y axis column

        color : str, optional
            Column name to add a dimension by color.

        orient : str, optional
            Orientation of graph, 'h' for horizontal
            'v' for vertical, by default 'v',

        points : str, bool {'outlier', 'suspectedoutliers', 'all', False}
            One of 'outliers', 'suspectedoutliers', 'all', or False.
            If 'outliers', only the sample points lying outside the whiskers are shown.
            If 'suspectedoutliers', all outlier points are shown and those less than 4*Q1-3*Q3 or greater than 4*Q3-3*Q1 are highlighted with the marker’s 'outliercolor'.
            If 'outliers', only the sample points lying outside the whiskers are shown.
            If 'all', all sample points are shown. If False, no sample points are shown and the whiskers extend to the full range of the sample.

        notched : bool, optional
            If True, boxes are drawn with notches, by default False.

        title : str, optional
            Title of the plot, by default "".

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.boxplot(y='y', color='z')
        >>> data.boxplot(x='x', y='y', color='z', points='all')
        >>> data.boxplot(x='x', y='y', output_file='pair.png')
        """

        assert (x is not None or y is not None), "An x column or a y column must be provided."

        fig = viz.boxplot(
            x=x,
            y=y,
            data=self.train_data,
            color=color,
            title=title,
            output_file=output_file,
            **kwargs
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig

    def violinplot(self, x=None, y=None, color=None, title="", output_file="", **kwargs):
        """
        Plots a violin plot for the given x and y columns.

        For more info and kwargs for violin plots, see https://plot.ly/python-api-reference/generated/plotly.express.violin.html#plotly.express.violin 
        and https://plot.ly/python/violin/
        
        Parameters
        ----------
        x : str
            X axis column

        y : str
            y axis column

        color : str, optional
            Column name to add a dimension by color.

        orient : str, optional
            Orientation of graph, 'h' for horizontal
            'v' for vertical, by default 'v',

        points : str, bool {'outlier', 'suspectedoutliers', 'all', False}
            One of 'outliers', 'suspectedoutliers', 'all', or False.
            If 'outliers', only the sample points lying outside the whiskers are shown.
            If 'suspectedoutliers', all outlier points are shown and those less than 4*Q1-3*Q3 or greater than 4*Q3-3*Q1 are highlighted with the marker’s 'outliercolor'.
            If 'outliers', only the sample points lying outside the whiskers are shown.
            If 'all', all sample points are shown. If False, no sample points are shown and the whiskers extend to the full range of the sample.

        violinmode : str {'group', 'overlay'}
            In 'overlay' mode, violins are on drawn top of one another.
            In 'group' mode, violins are placed beside each other.

        box : bool, optional
            If True, boxes are drawn inside the violins.

        title : str, optional
            Title of the plot, by default "".

        output_file : str, optional
            Output file name for image with extension (i.e. jpeg, png, etc.)

        Examples
        --------
        >>> data.violinplot(y='y', color='z', box=True)
        >>> data.violinplot(x='x', y='y', color='z', points='all')
        >>> data.violinplot(x='x', y='y', violinmode='overlay', output_file='pair.png')
        """

        assert (x is not None or y is not None), "An x column or a y column must be provided."

        fig = viz.violinplot(
            x=x,
            y=y,
            data=self.train_data,
            color=color,
            title=title,
            output_file=output_file,
            **kwargs
        )

        if output_file and self.report:  # pragma: no cover
            self.report.write_image(output_file)

        return fig



import scipy as sc
from aethos.visualizations import visualize as viz


def run_2sample_ttest(
    group1: str, group2: str, train_data, t_type: str, output_file, **kwargs
):
    """
    Helper function to run 2 sample ttest.
    
    Parameters
    ----------
    group1 : str
        Column for group 1 to compare.
    
    group2 : str, optional
        Column for group 2 to compare, by default None

    train_data : DataFrame
        Train Data

    t_type : str
        T-test type

    output_file : str
        File name to output
    
    Returns
    -------
    list
        T test statistic, P value        
    """

    data_group1 = train_data[group1].tolist()
    data_group2 = train_data[group2].tolist()

    if t_type == "ind":
        results = sc.stats.ttest_ind(
            data_group1, data_group2, nan_policy="omit", **kwargs
        )
    else:
        results = sc.stats.ttest_rel(data_group1, data_group2, nan_policy="omit",)

    matrix = [
        ["", "Test Statistic", "p-value"],
        ["Sample Data", results[0], results[1]],
    ]

    viz.create_table(matrix, True, output_file)

    return results



import matplotlib.pyplot as plt
import numpy as np
import itertools
import pandas as pd
import scipy as sc
from aethos.config import technique_reason_repo
from scipy.stats.stats import ks_2samp
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.metrics import classification_report
from sklearn.model_selection import StratifiedKFold, cross_val_predict
from collections import Counter
from typing import Union
from aethos.stats.util import run_2sample_ttest


class Stats(object):
    def predict_data_sample(self):
        """
        Identifies how similar the train and test set distribution are by trying to predict whether each sample belongs
        to the train or test set using Random Forest, 10 Fold Stratified Cross Validation.

        The lower the F1 score, the more similar the distributions are as it's harder to predict which sample belongs to which distribution.

        Credit: https://www.kaggle.com/nanomathias/distribution-of-test-vs-training-data#1.-t-SNE-Distribution-Overview

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.predict_data_sample()
        """

        if self.x_test is None or not self.target_field:
            raise ValueError(
                "Test data or target field must be set. They can be set by assigning values to the `target_field` or the `x_test` variable."
            )

        report_info = technique_reason_repo["stats"]["dist_compare"]["predict"]

        x_train = self.x_train.drop(self.target_field, axis=1)
        x_test = self.x_test.drop(self.target_field, axis=1)

        x_train["label"] = 1
        x_test["label"] = 0

        data = pd.concat([x_train, x_test], axis=0)
        label = data["label"].tolist()

        predictions = cross_val_predict(
            ExtraTreesClassifier(n_estimators=100),
            data.drop(columns=["label"]),
            label,
            cv=StratifiedKFold(n_splits=10, shuffle=True, random_state=42),
        )

        if self.report is not None:
            self.report.report_technique(report_info, [])

        print(classification_report(data["label"].tolist(), predictions))

        return self.copy()

    def ks_feature_distribution(self, threshold=0.1, show_plots=True):
        """
        Uses the Kolomogorov-Smirnov test see if the distribution in the training and test sets are similar.
        
        Credit: https://www.kaggle.com/nanomathias/distribution-of-test-vs-training-data#1.-t-SNE-Distribution-Overview

        Parameters
        ----------
        threshold : float, optional
            KS statistic threshold, by default 0.1

        show_plots : bool, optional
            True to show histograms of feature distributions, by default True

        Returns
        -------
        DataFrame
            Columns that are significantly different in the train and test set.

        Examples
        --------
        >>> data.ks_feature_distribution()
        >>> data.ks_feature_distribution(threshold=0.2)
        """

        import swifter
        from tqdm import tqdm

        if self.x_test is None:
            raise ValueError(
                "Data must be split into train and test set. Please set the `x_test` variable."
            )

        report_info = technique_reason_repo["stats"]["dist_compare"]["ks"]

        diff_data = []
        diff_df = None

        for col in tqdm(self.x_train.columns):
            statistic, pvalue = ks_2samp(
                self.x_train[col].values, self.x_test[col].values
            )

            if pvalue <= 0.05 and np.abs(statistic) > threshold:
                diff_data.append(
                    {
                        "feature": col,
                        "p": np.round(pvalue, 5),
                        "statistic": np.round(np.abs(statistic), 2),
                    }
                )

        if diff_data:
            diff_df = pd.DataFrame(diff_data).sort_values(
                by=["statistic"], ascending=False
            )

            if show_plots:
                n_cols = 4
                n_rows = int(len(diff_df) / n_cols) + 1

                _, ax = plt.subplots(n_rows, n_cols, figsize=(40, 8 * n_rows))

                for i, (_, row) in enumerate(diff_df.iterrows()):
                    if i >= len(ax):
                        break

                    extreme = np.max(
                        np.abs(
                            self.x_train[row.feature].tolist()
                            + self.x_test[row.feature].tolist()
                        )
                    )
                    self.x_train.loc[:, row.feature].swifter.apply(np.log1p).hist(
                        ax=ax[i],
                        alpha=0.6,
                        label="Train",
                        density=True,
                        bins=np.arange(-extreme, extreme, 0.25),
                    )

                    self.x_test.loc[:, row.feature].swifter.apply(np.log1p).hist(
                        ax=ax[i],
                        alpha=0.6,
                        label="Train",
                        density=True,
                        bins=np.arange(-extreme, extreme, 0.25),
                    )

                    ax[i].set_title(f"Statistic = {row.statistic}, p = {row.p}")
                    ax[i].set_xlabel(f"Log({row.feature})")
                    ax[i].legend()

                plt.tight_layout()
                plt.show()

            if self.report is not None:
                self.report.report_technique(report_info, [])

        return diff_df

    def most_common(
        self, col: str, n=15, plot=False, use_test=False, output_file="", **plot_kwargs
    ):
        """
        Analyzes the most common values in the column and either prints them or displays a bar chart.
        
        Parameters
        ----------
        col : str
            Column to analyze

        n : int, optional
            Number of top most common values to display, by default 15

        plot : bool, optional
            True to plot a bar chart, by default False

        use_test : bool, optional
            True to analyze the test set, by default False

        output_file : str,
            File name to save plot as, IF plot=True

        Examples
        --------
        >>> data.most_common('col1', plot=True)
        >>> data.most_common('col1', n=50, plot=True)
        >>> data.most_common('col1', n=50)
        """

        if use_test:
            data = self.x_test[col].tolist()
        else:
            data = self.x_train[col].tolist()

        test_sample = data[0]

        if isinstance(test_sample, list):
            data = itertools.chain(*map(list, data))
        elif isinstance(test_sample, str):
            data = map(str.split, data)
            data = itertools.chain(*data)

        counter = Counter(data)
        most_common = dict(counter.most_common(n))

        if plot:
            from aethos.visualizations.visualize import barplot

            df = pd.DataFrame(list(most_common.items()), columns=["Word", "Count"])

            fig = barplot(
                x="Word", y="Count", data=df, output_file=output_file, **plot_kwargs
            )

            return fig
        else:
            for k, v in most_common.items():
                print(f"{k}: {v}")

            return most_common

    def ind_ttest(self, group1: str, group2: str, equal_var=True, output_file=None):
        """
        Performs an Independent T test.

        This is to be used when you want to compare the means of 2 groups.

        If group 2 column name is not provided and there is a test set, it will compare the same column in the train and test set.

        If there are any NaN's they will be omitted.
        
        Parameters
        ----------
        group1 : str
            Column for group 1 to compare.
        
        group2 : str, optional
            Column for group 2 to compare, by default None

        equal_var : bool, optional
            If True (default), perform a standard independent 2 sample test that assumes equal population variances.
            If False, perform Welch's t-test, which does not assume equal population variance, by default True

        output_file : str, optional
            Name of the file to output, by default None

        Returns
        -------
        list
            T test statistic, P value

        Examples
        --------
        >>> data.ind_ttest('col1', 'col2')
        >>> data.ind_ttest('col1', 'col2', output_file='ind_ttest.png')
        """

        res = run_2sample_ttest(
            group1, group2, self.x_train, "ind", output_file, equal_var=equal_var
        )

        return res

    def paired_ttest(self, group1: str, group2=None, output_file=None):
        """
        Performs a Paired t-test.

        This is to be used when you want to compare the means from the same group at different times.

        If group 2 column name is not provided and there is a test set, it will compare the same column in the train and test set.

        If there are any NaN's they will be omitted.
        
        Parameters
        ----------
        group1 : str
            Column for group 1 to compare.
        
        group2 : str, optional
            Column for group 2 to compare, by default None

        equal_var : bool, optional
            If True (default), perform a standard independent 2 sample test that assumes equal population variances.
            If False, perform Welch's t-test, which does not assume equal population variance, by default True

        output_file : str, optional
            Name of the file to output, by default None

        Returns
        -------
        list
            T test statistic, P value

        Examples
        --------
        >>> data.paired_ttest('col1', 'col2')
        >>> data.paired_ttest('col1', 'col2', output_file='pair_ttest.png')
        """

        # The implementation is the same as an independent t-test
        res = run_2sample_ttest(group1, group2, self.x_train, "pair", output_file)

        return res

    def onesample_ttest(self, group1: str, mean: Union[float, int], output_file=None):
        """
        Performs a One Sample t-test.

        This is to be used when you want to compare the mean of a single group against a known mean.

        If there are any NaN's they will be omitted.
        
        Parameters
        ----------
        group1 : str
            Column for group 1 to compare.
        
        mean : float, int, optional
            Sample mean to compare to.

        output_file : str, optional
            Name of the file to output, by default None

        Returns
        -------
        list
            T test statistic, P value

        Examples
        --------
        >>> data.onesample_ttest('col1', 1)
        >>> data.onesample_ttest('col1', 1, output_file='ones_ttest.png')
        """

        from aethos.visualizations import visualize as viz

        data_group1 = self.x_train[group1].tolist()

        results = sc.stats.ttest_1samp(data_group1, mean, nan_policy="omit")

        matrix = [
            ["", "Test Statistic", "p-value"],
            ["Sample Data", results[0], results[1]],
        ]

        viz.create_table(matrix, True, output_file)

        return results

    def anova(
        self,
        dep_var: str,
        num_variables=[],
        cat_variables=[],
        formula=None,
        verbose=False,
    ):
        """
        Runs an anova.

        Anovas are to be used when one wants to compare the means of a condition between 2+ groups.

        ANOVA tests if there is a difference in the mean somewhere in the model (testing if there was an overall effect), but it does not tell one where the difference is if the there is one.
        
        Parameters
        ----------
        dep_var : str
            Dependent variable you want to explore the relationship of

        num_variables : list, optional
            Numeric variable columns, by default []

        cat_variables : list, optional
            Categorical variable columns, by default []

        formula : str, optional
            OLS formula statsmodel lib, by default None

        verbose : bool, optional
            True to print OLS model summary and formula, by default False

        Returns
        -------


        Examples
        --------
        >>> data.anova('dep_col', num_variables=['col1', 'col2'], verbose=True)
        >>> data.anova('dep_col', cat_variables=['col1', 'col2'], verbose=True)
        >>> data.anova('dep_col', num_variables=['col1', 'col2'], cat_variables=['col3'] verbose=True)
        """

        from statsmodels.formula.api import ols
        import statsmodels.api as sm

        assert (
            num_variables != [] or cat_variables != []
        ), "You must specify variables, either categorical or numerical."

        # Create the formula string to pass into OLS in the form of `dep_colname` ~ `num_col1` + C(`cat_col1`) + ...
        cat_variables = [f"C({var})" for var in cat_variables]
        join = "+" if cat_variables and num_variables else ""
        formula = (
            f'{dep_var} ~ {" + ".join(num_variables)} {join} {" + ".join(cat_variables)}'
            if not formula
            else formula
        )

        mod = ols(formula, data=self.x_train).fit()

        if verbose:
            print(formula)
            print(mod.summary())

        table = sm.stats.anova_lm(mod, typ=2)

        print(table)



from aethos.cleaning import util
from aethos.cleaning import categorical as cat
from aethos.cleaning import numeric as num
from aethos.config import technique_reason_repo
from aethos.util import _input_columns, _numeric_input_conditions


class Clean(object):
    def drop_column_missing_threshold(self, threshold: float):
        """
        Remove columns from the dataframe that have greater than or equal to the threshold value of missing values.
        Example: Remove columns where >= 50% of the data is missing.

        This function exists in `clean/utils.py`
        
        Parameters
        ----------
        threshold : float
            Value between 0 and 1 that describes what percentage of a column can be missing values.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_column_missing_threshold(0.5)
        """

        report_info = technique_reason_repo["clean"]["general"]["remove_columns"]

        original_columns = set(list(self.x_train.columns))

        (self.x_train, self.x_test,) = util.remove_columns_threshold(
            x_train=self.x_train, x_test=self.x_test, threshold=threshold,
        )

        if self.report is not None:
            new_columns = original_columns.difference(self.x_train.columns)
            self.report.report_technique(report_info, new_columns)

        return self.copy()

    def drop_constant_columns(self):
        """
        Remove columns from the data that only have one unique value.

        This function exists in `clean/utils.py`
                
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_constant_columns()
        """

        report_info = technique_reason_repo["clean"]["general"][
            "remove_constant_columns"
        ]

        original_columns = set(list(self.x_train.columns))

        (self.x_train, self.x_test,) = util.remove_constant_columns(
            x_train=self.x_train, x_test=self.x_test
        )

        if self.report is not None:
            new_columns = original_columns.difference(self.x_train.columns)
            self.report.report_technique(report_info, new_columns)

        return self.copy()

    def drop_unique_columns(self):
        """
        Remove columns from the data that only have one unique value.

        This function exists in `clean/utils.py`
                
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_unique_columns()
        """

        report_info = technique_reason_repo["clean"]["general"]["remove_unique_columns"]

        original_columns = set(list(self.x_train.columns))

        (self.x_train, self.x_test,) = util.remove_unique_columns(
            x_train=self.x_train, x_test=self.x_test
        )

        if self.report is not None:
            new_columns = original_columns.difference(self.x_train.columns)
            self.report.report_technique(report_info, new_columns)

        return self.copy()

    def drop_rows_missing_threshold(self, threshold: float):
        """
        Remove rows from the dataframe that have greater than or equal to the threshold value of missing rows.
        Example: Remove rows where > 50% of the data is missing.

        This function exists in `clean/utils.py`.

        Parameters
        ----------
        threshold : float
            Value between 0 and 1 that describes what percentage of a row can be missing values.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_rows_missing_threshold(0.5)    
        """

        report_info = technique_reason_repo["clean"]["general"]["remove_rows"]

        (self.x_train, self.x_test,) = util.remove_rows_threshold(
            x_train=self.x_train, x_test=self.x_test, threshold=threshold,
        )

        # Write to report
        if self.report is not None:
            self.report.report_technique(report_info)

        return self.copy()

    def replace_missing_mean(self, *list_args, list_of_cols=[]):
        """
        Replaces missing values in every numeric column with the mean of that column.

        If no columns are supplied, missing values will be replaced with the mean in every numeric column.

        Mean: Average value of the column. Effected by outliers.

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/numeric.py` as `replace_missing_mean_median_mode`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to

        list_of_cols : list, optional
            Specific columns to apply this technique to, by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_mean('col1', 'col2')
        >>> data.replace_missing_mean(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["numeric"]["mean"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = num.replace_missing_mean_median_mode(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            strategy="mean",
        )

        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_median(self, *list_args, list_of_cols=[]):
        """
        Replaces missing values in every numeric column with the median of that column.

        If no columns are supplied, missing values will be replaced with the mean in every numeric column.

        Median: Middle value of a list of numbers. Equal to the mean if data follows normal distribution. Not effected much by anomalies.

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/numeric.py` as `replace_missing_mean_median_mode`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            Specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_median('col1', 'col2')
        >>> data.replace_missing_median(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["numeric"]["median"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = num.replace_missing_mean_median_mode(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            strategy="median",
        )

        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_mostcommon(self, *list_args, list_of_cols=[]):
        """
        Replaces missing values in every numeric column with the most common value of that column

        Mode: Most common value.

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/numeric.py` as `replace_missing_mean_median_mode`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_mostcommon('col1', 'col2')
        >>> data.replace_missing_mostcommon(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["numeric"]["mode"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = num.replace_missing_mean_median_mode(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            strategy="most_frequent",
        )
        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_constant(
        self, *list_args, list_of_cols=[], constant=0, col_mapping=None
    ):
        """
        Replaces missing values in every numeric column with a constant.

        If no columns are supplied, missing values will be replaced with the mean in every numeric column.

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/numeric.py` as `replace_missing_constant`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        constant : int or float, optional
            Numeric value to replace all missing values with , by default 0

        col_mapping : dict, optional
            Dictionary mapping {'ColumnName': `constant`}, by default None
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_constant(col_mapping={'a': 1, 'b': 2, 'c': 3})
        >>> data.replace_missing_constant('col1', 'col2', constant=2)
        >>> data.replace_missing_constant(['col1', 'col2'], constant=3)
        """

        report_info = technique_reason_repo["clean"]["numeric"]["constant"]

        if col_mapping:
            col_to_constant = col_mapping
        else:
            # If a list of columns is provided use the list, otherwise use arguemnts.
            col_to_constant = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = num.replace_missing_constant(
            x_train=self.x_train,
            x_test=self.x_test,
            col_to_constant=col_to_constant,
            constant=constant,
        )

        if self.report is not None:
            if not col_to_constant:
                self.report.report_technique(report_info, self.x_train.columns)
            else:
                self.report.report_technique(report_info, list(col_to_constant))

        return self.copy()

    def replace_missing_new_category(
        self, *list_args, list_of_cols=[], new_category=None, col_mapping=None
    ):
        """
        Replaces missing values in categorical column with its own category. The categories can be autochosen
        from the defaults set.

        For numeric categorical columns default values are: -1, -999, -9999
        For string categorical columns default values are: "Other", "Unknown", "MissingDataCategory"

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/categorical.py` as `replace_missing_new_category`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_category : str, int, or float, optional
            Category to replace missing values with, by default None

        col_mapping : dict, optional
           Dictionary mapping {'ColumnName': `constant`}, by default None
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_new_category(col_mapping={'col1': "Green", 'col2': "Canada", 'col3': "December"})
        >>> data.replace_missing_new_category('col1', 'col2', 'col3', new_category='Blue')
        >>> data.replace_missing_new_category(['col1', 'col2', 'col3'], new_category='Blue')
        """

        report_info = technique_reason_repo["clean"]["categorical"]["new_category"]

        # If dictionary mapping is provided, use that otherwise use column
        if col_mapping:
            col_to_category = col_mapping
        else:
            # If a list of columns is provided use the list, otherwise use arguemnts.
            col_to_category = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = cat.replace_missing_new_category(
            x_train=self.x_train,
            x_test=self.x_test,
            col_to_category=col_to_category,
            constant=new_category,
        )

        if self.report is not None:
            if not col_to_category:
                self.report.report_technique(report_info, self.x_train.columns)
            else:
                self.report.report_technique(report_info, list(col_to_category))

        return self.copy()

    def replace_missing_remove_row(self, *list_args, list_of_cols=[]):
        """
        Remove rows where the value of a column for those rows is missing.

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/categorical.py` as `replace_missing_remove_row`.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_remove_row('col1', 'col2')
        >>> data.replace_missing_remove_row(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["categorical"]["remove_rows"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = cat.replace_missing_remove_row(
            x_train=self.x_train, x_test=self.x_test, cols_to_remove=list_of_cols,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def drop_duplicate_rows(self, *list_args, list_of_cols=[]):
        """
        Remove rows from the data that are exact duplicates of each other and leave only 1.
        This can be used to reduce processing time or performance for algorithms where
        duplicates have no effect on the outcome (i.e DBSCAN)

        If a list of columns is provided use the list, otherwise use arguemnts.

        This function exists in `clean/util.py` as `remove_duplicate_rows`.
       
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
       
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_duplicate_rows('col1', 'col2') # Only look at columns 1 and 2
        >>> data.drop_duplicate_rows(['col1', 'col2'])
        >>> data.drop_duplicate_rows()
        """

        report_info = technique_reason_repo["clean"]["general"]["remove_duplicate_rows"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.remove_duplicate_rows(
            x_train=self.x_train, x_test=self.x_test, list_of_cols=list_of_cols,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def drop_duplicate_columns(self):
        """
        Remove columns from the data that are exact duplicates of each other and leave only 1.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop_duplicate_columns()
        """

        report_info = technique_reason_repo["clean"]["general"][
            "remove_duplicate_columns"
        ]

        (self.x_train, self.x_test,) = util.remove_duplicate_columns(
            x_train=self.x_train, x_test=self.x_test
        )

        if self.report is not None:
            self.report.report_technique(report_info)

        return self.copy()

    def replace_missing_random_discrete(self, *list_args, list_of_cols=[]):
        """
        Replace missing values in with a random number based off the distribution (number of occurences) 
        of the data.

        For example if your data was [5, 5, NaN, 1, 2]
        There would be a 50% chance that the NaN would be replaced with a 5, a 25% chance for 1 and a 25% chance for 2.

        If a list of columns is provided use the list, otherwise use arguemnts.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.
            
        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_random_discrete('col1', 'col2')
        >>> data.replace_missing_random_discrete(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["general"]["random_discrete"]

        ## If a list of columns is provided use the list, otherwise use arguemnts.
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.replace_missing_random_discrete(
            x_train=self.x_train, x_test=self.x_test, list_of_cols=list_of_cols,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_knn(self, k=5, **knn_kwargs):
        """
        Replaces missing data with data from similar records based off a distance metric.

        For more info see: https://scikit-learn.org/stable/modules/generated/sklearn.impute.KNNImputer.html#sklearn.impute.KNNImputer
        
        Parameters
        ----------
        missing_values : number, string, np.nan or None, default=`np.nan`
            The placeholder for the missing values. All occurrences of missing_values will be imputed.

        k : int, default=5
            Number of neighboring samples to use for imputation.

        weights : {‘uniform’, ‘distance’} or callable, default=’uniform’
            Weight function used in prediction. Possible values:

                ‘uniform’ : uniform weights. All points in each neighborhood are weighted equally.

                ‘distance’ : weight points by the inverse of their distance. in this case, closer neighbors of a query point will have a greater influence than neighbors which are further away.

                callable : a user-defined function which accepts an array of distances, and returns an array of the same shape containing the weights.

        metric : {‘nan_euclidean’} or callable, default=’nan_euclidean’
            Distance metric for searching neighbors. Possible values:

                ‘nan_euclidean’

                callable : a user-defined function which conforms to the definition of _pairwise_callable(X, Y, metric, **kwds).
                The function accepts two arrays, X and Y, and a missing_values keyword in kwds and returns a scalar distance value.

        add_indicator : bool, default=False
            If True, a MissingIndicator transform will stack onto the output of the imputer’s transform.
            This allows a predictive estimator to account for missingness despite imputation.
            If a feature has no missing values at fit/train time, the feature won’t appear on the missing indicator even if there are missing values at transform/test time.

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_knn(k=8)
        """

        report_info = technique_reason_repo["clean"]["general"]["knn"]

        (self.x_train, self.x_test,) = util.replace_missing_knn(
            x_train=self.x_train, x_test=self.x_test, n_neighbors=k, **knn_kwargs
        )

        if self.report is not None:
            self.report.report_technique(report_info)

        return self.copy()

    def replace_missing_interpolate(
        self, *list_args, list_of_cols=[], method="linear", **inter_kwargs
    ):
        """
        Replaces missing values with an interpolation method and possible extrapolation.

        The possible interpolation methods are:
           
            - 'linear': Ignore the index and treat the values as equally spaced. This is the only method supported on MultiIndexes.
            - 'time': Works on daily and higher resolution data to interpolate given length of interval.
            - 'index', ‘values’: use the actual numerical values of the index.
            - 'pad': Fill in NaNs using existing values.
            - 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'spline', ‘barycentric’, ‘polynomial’: Passed to scipy.interpolate.interp1d.
                - These methods use the numerical values of the index. Both ‘polynomial’ and ‘spline’ require that you also specify an order (int), e.g. df.interpolate(method='polynomial', order=5).
            - 'krogh', 'piecewise_polynomial', 'spline', 'pchip', 'akima': Wrappers around the SciPy interpolation methods of similar names.
            - 'from_derivatives': Refers to scipy.interpolate.BPoly.from_derivatives which replaces ‘piecewise_polynomial’ interpolation method in scipy 0.18.

        For more information see: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.interpolate.html or https://docs.scipy.org/doc/scipy/reference/interpolate.html.

        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.
            
        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        method : str, optional
            Interpolation method, by default 'linear'

        limit : int, optional
            Maximum number of consecutive NaNs to fill. Must be greater than 0.

        limit_area : {None, ‘inside’, ‘outside’}, default None
            If limit is specified, consecutive NaNs will be filled with this restriction.

            - None: No fill restriction.
            - ‘inside’: Only fill NaNs surrounded by valid values (interpolate).
            - ‘outside’: Only fill NaNs outside valid values (extrapolate).
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_interpolate('col1', 'col2')
        >>> data.replace_missing_interpolate(['col1', 'col2'])
        >>> data.replace_missing_interpolate('col1', 'col2', method='pad', limit=3)
        """

        report_info = technique_reason_repo["clean"]["general"]["interpolate"]
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.replace_missing_interpolate(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            method=method,
            **inter_kwargs
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_backfill(self, *list_args, list_of_cols=[], **extra_kwargs):
        """
        Replaces missing values in a column with the next known data point.

        This is useful when dealing with timeseries data and you want to replace data in the past with data from the future.

        For more info view the following link: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.fillna.html
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.
            
        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_backfill('col1', 'col2')
        >>> data.replace_missing_backfill(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["general"]["bfill"]
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.replace_missing_fill(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            method="bfill",
            **extra_kwargs
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_forwardfill(self, *list_args, list_of_cols=[], **extra_kwargs):
        """
        Replaces missing values in a column with the last known data point.

        This is useful when dealing with timeseries data and you want to replace future missing data with the past.

        For more info view the following link: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.fillna.html
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.
            
        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_forwardfill('col1', 'col2')
        >>> data.replace_missing_forwardfill(['col1', 'col2'])
        """

        report_info = technique_reason_repo["clean"]["general"]["ffill"]
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.replace_missing_fill(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            method="ffill",
            **extra_kwargs
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def replace_missing_indicator(
        self,
        *list_args,
        list_of_cols=[],
        missing_indicator=1,
        valid_indicator=0,
        keep_col=True
    ):
        """
        Adds a new column describing whether data is missing for each record in a column.

        This is useful if the missing data has meaning, aka not random.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.
            
        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        missing_indicator : int, optional
            Value to indicate missing data, by default 1

        valid_indicator : int, optional
            Value to indicate non missing data, by default 0

        keep_col : bool, optional
            True to keep column, False to replace it, by default False
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.replace_missing_indicator('col1', 'col2')
        >>> data.replace_missing_indicator(['col1', 'col2'])
        >>> data.replace_missing_indicator(['col1', 'col2'], missing_indicator='missing', valid_indicator='not missing', keep_col=False)
        """

        report_info = technique_reason_repo["clean"]["general"]["indicator"]
        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = util.replace_missing_indicator(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            missing_indicator=missing_indicator,
            valid_indicator=valid_indicator,
            keep_col=keep_col,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()



import inspect
import multiprocessing as mp
import os
import pickle
import warnings
from functools import partial, wraps
from pathlib import Path

import catboost as cb
import lightgbm as lgb
import matplotlib.pyplot as plt

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=DeprecationWarning)

    import mlflow
    import mlflow.lightgbm
    import mlflow.sklearn
    import mlflow.xgboost

import xgboost as xgb
from sklearn.model_selection import GridSearchCV, KFold, StratifiedKFold
from yellowbrick.model_selection import CVScores, LearningCurve

from aethos.config import EXP_DIR, DEFAULT_MODEL_DIR, IMAGE_DIR, cfg
from aethos.config.config import _global_config
from aethos.util import _make_dir
from pickle import dump


def add_to_queue(model_function):
    @wraps(model_function)
    def wrapper(self, *args, **kwargs):
        default_kwargs = get_default_args(model_function)

        kwargs["run"] = kwargs.get("run", default_kwargs["run"])
        kwargs["model_name"] = kwargs.get("model_name", default_kwargs["model_name"])
        cv = kwargs.get("cv", False)

        if not _validate_model_name(self, kwargs["model_name"]):
            raise AttributeError("Invalid model name. Please choose another one.")

        if kwargs["run"]:
            return model_function(self, *args, **kwargs)
        else:
            if "XGBoost" in model_function.__name__:
                print(
                    "XGBoost is not comptabile to be run in parallel. Please run it normally or in run all models in a series."
                )
            else:
                kwargs["run"] = True
                self._queued_models[kwargs["model_name"]] = partial(
                    getattr(self, model_function.__name__), *args, **kwargs
                )

    return wrapper


def get_default_args(func):
    """
    Gets default arguments from a function.
    """

    signature = inspect.signature(func)

    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def run_gridsearch(model, gridsearch, cv=5, scoring="accuracy", **gridsearch_kwargs):
    """
    Runs Gridsearch on a model
    
    Parameters
    ----------
    model : Model
        Model to run gridsearch on

    gridsearch : dict
        Dict of params to test

    cv : int, Crossvalidation Generator, optional
        Cross validation method, by default 12
    
    scoring : str
        Scoring metric to use when evaluating models
    
    Returns
    -------
    Model
        Initialized Gridsearch model
    """

    if isinstance(gridsearch, dict):
        gridsearch_grid = gridsearch
        print(f"Gridsearching with the following parameters: {gridsearch_grid}")
    else:
        raise ValueError("Invalid Gridsearch input.")

    model = GridSearchCV(
        model, gridsearch_grid, cv=cv, scoring=scoring, **gridsearch_kwargs
    )

    return model


def run_crossvalidation(
    model, x_train, y_train, cv=5, scoring="accuracy", report=None, model_name=None
):
    """
    Runs cross validation on a certain model.
    
    Parameters
    ----------
    model : Model
        Model to cross validate

    x_train : nd-array
        Training data

    y_train : nd-array
        Testing data

    cv : int, Crossvalidation Generator, optional
        Cross validation method, by default 5

    scoring : str, optional
        Scoring method, by default 'accuracy'
    """

    fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(15, 5))
    visualizer_scores = CVScores(model, cv=cv, scoring=scoring, ax=axes[0])
    visualizer_scores.fit(x_train, y_train)
    visualizer_scores.finalize()

    visualizer_lcurve = LearningCurve(model, cv=cv, scoring=scoring, ax=axes[1])
    visualizer_lcurve.fit(x_train, y_train)
    visualizer_lcurve.finalize()

    visualizer_scores.show()
    visualizer_lcurve.show()

    if report or _global_config["track_experiments"]:  # pragma: no cover
        fig.savefig(os.path.join(IMAGE_DIR, model_name, "cv.png"))


def _run_models_parallel(model_obj):
    """
    Runs queued models in parallel
    
    Parameters
    ----------
    model_obj : Model
        Model object
    """

    with mp.Pool(mp.cpu_count()) as pool:
        results = pool.map(_run, list(model_obj._queued_models.values()))

    for result in results:
        model_obj._models[result.model_name] = result
        del model_obj._queued_models[result.model_name]

    return results


def _run(model):
    """
    Runs a model
        
    Returns
    -------
    Model
        Trained model
    """

    return model()


def _get_cv_type(cv_type, random_state, **kwargs):
    """
    Takes in cv type from the user and initiates the cross validation generator.
    
    Parameters
    ----------
    cv_type : int, str or None
        Crossvalidation type

    random_state : int
        Random seed
    
    Returns
    -------
    Cross Validation Generator
        CV Generator
    """

    if not cv_type:
        return None, kwargs

    if isinstance(cv_type, int):
        cv_type = cv_type
    elif cv_type == "kfold":
        n_splits = kwargs.pop("n_splits", 5)
        shuffle = kwargs.pop("shuffle", False)

        cv_type = KFold(n_splits=n_splits, shuffle=shuffle, random_state=random_state)
    elif cv_type == "strat-kfold":
        n_splits = kwargs.pop("n_splits", 5)
        shuffle = kwargs.pop("shuffle", False)

        cv_type = StratifiedKFold(
            n_splits=n_splits, shuffle=shuffle, random_state=random_state
        )
    else:
        raise ValueError("Cross Validation type is invalid.")

    return cv_type, kwargs


def to_pickle(model, name, project=False, project_name=None):
    """
    Writes model to a pickle file.
    
    Parameters
    ----------
    model: Model object
        Model object to serialize

    name : str
        Name of the model
        
    project : bool
        Whether to write to the project folder, by default False
    """

    if not project:
        if not cfg["models"]["dir"]:  # pragma: no cover
            path = DEFAULT_MODEL_DIR
        else:
            path = cfg["models"]["dir"]
    else:
        path = os.path.join(
            os.path.expanduser("~"), ".aethos", "projects", project_name, "app"
        )

    _make_dir(path)

    pickle.dump(model, open(os.path.join(path, name + ".pkl"), "wb"))


def _make_img_project_dir(model_name: str):
    """
    Make a model dir in images directory.
    
    Parameters
    ----------
    model_name : str
        Model name
    """

    _make_dir(os.path.join(IMAGE_DIR, model_name))


def _validate_model_name(model_obj, model_name: str) -> bool:
    """
    Validates the inputted model name. If the object already has an
    attribute with that model name, it is invalid
    
    Parameters
    ----------
    model_name : str
        Proposed name of the model
        
    model_obj : Model
        Model object
    
    Returns
    -------
    bool
        True if model name is valid, false otherwise
    """

    if hasattr(model_obj, model_name) and model_name not in model_obj._models:
        return False

    return True


def track_model(
    exp_name: str, model, model_name: str, model_kwargs: dict, metrics=None
):  # pragma: no cover
    """
    Logs model information into MLFlow console.
    
    Parameters
    ----------
    exp_name: str
        Name of the experiment

    model : Sklearn, LGBM, XGB, or CB Model object
        Model

    model_name : str
        Name of the model

    model_kwargs : dict
        Model parameters

    metrics : dict
        Metrics for the model
    """

    mlflow.set_tracking_uri(EXP_DIR)
    mlflow.set_experiment(exp_name)

    with mlflow.start_run(run_name=model_name) as run:

        mlflow.log_params(model_kwargs)
        mlflow.set_tag("name", model_name)

        if isinstance(model, xgb.XGBModel):
            mlflow.xgboost.log_model(model, model_name)
        else:
            mlflow.sklearn.log_model(model, model_name)

        mlflow.log_metrics(metrics)
        mlflow.log_artifacts(os.path.join(IMAGE_DIR, model_name))

        run_id = run.info.run_uuid

        mlflow.end_run()

    return run_id


def track_artifacts(run_id, model_name):  # pragma: no cover
    """
    Track artificats for modelling.
    
    Parameters
    ----------
    run_id : str
        MLFlow run id

    model_name : str
        Name of the model
    """

    with mlflow.start_run(run_id=run_id) as run:

        mlflow.log_artifacts(os.path.join(IMAGE_DIR, model_name))
        mlflow.end_run()



"""
This function has the following methods:

preprocess_normalize
"""

import numpy as np
import pandas as pd
from aethos.util import _numeric_input_conditions, drop_replace_columns
from sklearn.preprocessing import MinMaxScaler, RobustScaler

SCALER = {"minmax": MinMaxScaler, "robust": RobustScaler}


def scale(
    x_train,
    x_test=None,
    list_of_cols=[],
    method="minmax",
    keep_col=False,
    **algo_kwargs
):
    """
    Scales data according to a specific method.

    Parameters
    ----------
    x_train : DataFrame
        Dataset
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to
        If `list_of_cols` is not provided, the strategy will be
        applied to all numeric columns, by default []

    method : str, optional
        Scaling method, by default 'minmax'

    keep_col : bool, optional
        True to not remove the columns, by default False

    algo_kwargs : optional
        Parmaters to pass into the scaler constructor
        from Scikit-Learn, by default {}
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with rows normalized.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _numeric_input_conditions(list_of_cols, x_train)
    scaler = SCALER[method](**algo_kwargs)

    scaled_data = scaler.fit_transform(x_train[list_of_cols])
    scaled_df = pd.DataFrame(scaled_data, columns=list_of_cols)
    x_train = drop_replace_columns(x_train, list_of_cols, scaled_df, keep_col=keep_col)

    if x_test is not None:
        scaled_x_test = scaler.transform(x_test[list_of_cols])
        scaled_test_df = pd.DataFrame(scaled_x_test, columns=list_of_cols)
        x_test = drop_replace_columns(
            x_test, list_of_cols, scaled_test_df, keep_col=keep_col
        )

    return x_train, x_test


def log_scale(x_train, x_test=None, list_of_cols=[], base=None):
    """
    Scales data logarithmically.

    Options are '' for natural log, 2 for base2, 10 for base10.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset
        
    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list, optional
        A list of specific columns to apply this technique to
        If `list_of_cols` is not provided, the strategy will be
        applied to all numeric columns, by default []

    base : str, optional
        Base to logarithmically scale by, by default None
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with rows normalized.

    Returns 2 Dataframes if x_test is provided. 
    """

    list_of_cols = _numeric_input_conditions(list_of_cols, x_train)

    if not base:
        log = np.log
    elif base == 2:
        log = np.log2
    elif base == 10:
        log = np.log10
    else:
        log = np.log

    for col in list_of_cols:
        x_train[col] = log(x_train[col])

        if x_test is not None:
            x_test[col] = log(x_test[col])

    return x_train, x_test



import os

import click
import subprocess
from aethos.config.user_config import EXP_DIR


@click.group()
def main():
    pass


@main.command()
def create():
    """
    Creates a Data Science folder structure and files.
    """
    os.system("cookiecutter https://github.com/drivendata/cookiecutter-data-science")


@main.command()
def enable_extensions():
    """
    Enables jupyter extensions such as qgrid.
    """
    os.system("jupyter nbextension enable --py --sys-prefix widgetsnbextension")
    os.system("jupyter nbextension enable --py --sys-prefix qgrid")


@main.command()
def install_corpora():
    """
    Installs the necessary corpora from spaCy and NLTK for NLP analysis.
    """
    
    py_exe = ""

    try:
        if "3." in subprocess.check_output("python --version", stderr=subprocess.STDOUT, shell=True).decode("utf-8"):
            py_exe = "python"
    except subprocess.CalledProcessError as e:
        pass

    try:
        subprocess.check_output("python3 --version", stderr=subprocess.STDOUT, shell=True).decode("utf-8")
        py_exe = "python3"
    except  subprocess.CalledProcessError as e:
        pass
    
    if not py_exe:
        raise EnvironmentError("Python is not in your path, please add it your path.")

    os.system(f"{py_exe} -m textblob.download_corpora")
    os.system(f"{py_exe} -c 'import nltk; nltk.download(\"stopwords\")'")
    os.system(f"{py_exe} -m spacy download en")
    

@main.command()
@click.option(
    "-h",
    "--host",
    show_default=True,
    default="0.0.0.0",
    help="IP to bind to, default 0.0.0.0.",
)
@click.option(
    "-p",
    "--port",
    show_default=True,
    default="10000",
    help="Port to bind to, default 10000.",
)
def mlflow_ui(host, port):
    """
    Starts the MLFlow UI locally. If you are running MLFlow remotely, please start it there.
    """

    if not EXP_DIR.startswith("file:"):
        click.echo(
            "If you are running MLFlow remotely, please start it on the remote server."
        )
        click.echo(
            "If you are trying to run MLFlow locally, please the path starts like `file:/`."
        )

    os.system(
        f"mlflow ui -h {host} -p {port} --backend-store-uri {EXP_DIR[5:]} --default-artifact-root {EXP_DIR[5:]}"
    )



"""
This file contains the following methods:

feature_one_hot_encode
"""

import pandas as pd
from sklearn.preprocessing import OneHotEncoder

from aethos.util import _get_columns, drop_replace_columns


def feature_one_hot_encode(
    x_train, x_test=None, list_of_cols=[], keep_col=True, **algo_kwargs
):
    """
    Creates a matrix of converted categorical columns into binary columns of ones and zeros.
    
    Parameters
    ----------
    x_train : DataFrame
        Dataset

    x_test : DataFrame
        Testing dataset, by default None

    list_of_cols : list
         A list of specific columns to apply this technique to.

    keep_col : bool
        A parameter to specify whether to drop the column being transformed, by default
        keep the column, True

    algo_kwargs : optional
        Parameters you would pass into Bag of Words constructor as a dictionary, by default {"handle_unknown": "ignore"}
    
    Returns
    -------
    Dataframe, *Dataframe
        Transformed dataframe with the new column.

    Returns 2 Dataframes if x_test is provided. 
    """

    enc = OneHotEncoder(handle_unknown="ignore", **algo_kwargs)
    list_of_cols = _get_columns(list_of_cols, x_train)

    enc_data = enc.fit_transform(x_train[list_of_cols]).toarray()
    enc_df = pd.DataFrame(enc_data, columns=enc.get_feature_names(list_of_cols))
    x_train = drop_replace_columns(x_train, list_of_cols, enc_df, keep_col)

    if x_test is not None:
        enc_x_test = enc.transform(x_test[list_of_cols]).toarray()
        enc_test_df = pd.DataFrame(
            enc_x_test, columns=enc.get_feature_names(list_of_cols)
        )
        x_test = drop_replace_columns(x_test, list_of_cols, enc_test_df, keep_col)

    return x_train, x_test



import copy
import os
import re

import ipywidgets as widgets
import numpy as np
import pandas as pd
from aethos.cleaning.clean import Clean
from aethos.config import shell
from aethos.feature_engineering.feature import Feature
from aethos.preprocessing.preprocess import Preprocess
from aethos.reporting.report import Report
from aethos.stats.stats import Stats
from aethos.util import (
    CLEANING_CHECKLIST,
    DATA_CHECKLIST,
    ISSUES_CHECKLIST,
    MULTI_ANALYSIS_CHECKLIST,
    PREPARATION_CHECKLIST,
    UNI_ANALYSIS_CHECKLIST,
    _get_columns,
    _set_item,
    label_encoder,
    split_data,
)
from aethos.visualizations.visualizations import Visualizations
from IPython import get_ipython
from IPython.display import HTML, display
from ipywidgets import Layout


class Data(Clean, Preprocess, Feature, Visualizations, Stats):
    """
    Data class thats run analytical techniques.

    Parameters
    -----------
    x_train: pd.DataFrame
        Training data or aethos data object

    x_test: pd.DataFrame
        Test data, by default None

    split: bool
        True to split your training data into a train set and a test set

    test_split_percentage: float
        Percentage of data to split train data into a train and test set.
        Only used if `split=True`

    target_field: str
        For supervised learning problems, the name of the column you're trying to predict.

    report_name: str
        Name of the report to generate, by default None
    """

    def __init__(
        self,
        x_train,
        x_test=None,
        split=True,
        test_split_percentage=0.2,
        target_field="",
        report_name=None,
    ):

        self.x_train = x_train
        self.x_test = x_test
        self.split = split
        self.target_field = target_field
        self.report_name = report_name
        self.test_split_percentage = test_split_percentage
        self.target_mapping = None

        if split and x_test is None:
            # Generate train set and test set.
            self.x_train, self.x_test = split_data(self.x_train, test_split_percentage)
            self.x_train.reset_index(drop=True, inplace=True)
            self.x_test.reset_index(drop=True, inplace=True)

        if report_name is not None:
            self.report = Report(report_name)
            self.report_name = self.report.filename
        else:
            self.report = None
            self.report_name = None

    def __repr__(self):

        return self.x_train.to_string()

    def _repr_html_(self):  # pragma: no cover

        return self.x_train._repr_html_()

    def __getitem__(self, column):

        try:
            return self.x_train[column]

        except Exception as e:
            raise AttributeError(e)

    def __setitem__(self, column, value):

        if not self.split:
            self.x_train[column] = value

            return self.x_train.head()
        else:
            x_train_length = self.x_train.shape[0]
            x_test_length = self.x_test.shape[0]

            if isinstance(value, (list, np.ndarray)):
                ## If the number of entries in the list does not match the number of rows in the training or testing
                ## set raise a value error
                if len(value) != x_train_length and len(value) != x_test_length:
                    raise ValueError(
                        f"Length of list: {str(len(value))} does not equal the number rows as the training set or test set."
                    )

                self.x_train, self.x_test = _set_item(
                    self.x_train,
                    self.x_test,
                    column,
                    value,
                    x_train_length,
                    x_test_length,
                )

            elif isinstance(value, tuple):
                for data in value:
                    if len(data) != x_train_length and len(data) != x_test_length:
                        raise ValueError(
                            f"Length of list: {str(len(value))} does not equal the number rows as the training set or test set."
                        )

                    (self.x_train, self.x_test,) = _set_item(
                        self.x_train,
                        self.x_test,
                        column,
                        data,
                        x_train_length,
                        x_test_length,
                    )

            else:
                self.x_train[column] = value
                self.x_test[column] = value

            return self.x_train.head()

    def __getattr__(self, key):

        if key in self.__dict__:
            return getattr(self, key)

        if key in self.x_train.columns:
            return self.x_train[key]
        else:
            if hasattr(self.x_train, key):
                return getattr(self.x_train, key)
            else:
                raise AttributeError(f"Data object does not have attribute {key}.")

    def __setattr__(self, item, value):

        if item not in self.__dict__ or hasattr(
            self, item
        ):  # any normal attributes are handled normally
            dict.__setattr__(self, item, value)
        else:
            self.__setitem__(item, value)

    def __deepcopy__(self, memo):

        new_inst = type(self)(
            x_train=self.x_train,
            x_test=self.x_test,
            split=self.split,
            test_split_percentage=self.test_split_percentage,
            target_field=self.target_field,
            report_name=self.report_name,
        )

        return new_inst

    @property
    def y_train(self):
        """
        Property function for the training predictor variable
        """

        return self.x_train[self.target_field] if self.target_field else None

    @y_train.setter
    def y_train(self, value):
        """
        Setter function for the training predictor variable
        """

        if self.target_field:
            self.x_train[self.target_field] = value
        else:
            self.target_field = "label"
            self.x_train["label"] = value
            print('Added a target (predictor) field (column) named "label".')

    @property
    def y_test(self):
        """
        Property function for the testing predictor variable
        """

        if self.x_test is not None:
            if self.target_field:
                return self.x_test[self.target_field]
            else:
                return None
        else:
            return None

    @y_test.setter
    def y_test(self, value):
        """
        Setter function for the testing predictor variable
        """

        if self.x_test is not None:
            if self.target_field:
                self.x_test[self.target_field] = value
            else:
                self.target_field = "label"
                self.x_test["label"] = value
                print('Added a target (predictor) field (column) named "label".')

    @y_test.setter
    def y_test(self, value):
        """
        Setter function for the testing predictor variable
        """

        if self.x_test is not None:
            if self.target_field:
                self.x_test[self.target_field] = value
            else:
                self.target_field = "label"
                self.x_test["label"] = value
                print('Added a target (predictor) field (column) named "label".')

    @property
    def columns(self):
        """
        Property to return columns in the dataset.
        """

        return self.x_train.columns.tolist()

    @property
    def missing_values(self):
        """
        Property function that shows how many values are missing in each column.
        """

        dataframes = list(
            filter(lambda x: x is not None, [self.x_train, self.x_test,],)
        )

        missing_df = []
        for ind, dataframe in enumerate(dataframes):
            caption = (
                "Train set missing values." if ind == 0 else "Test set missing values."
            )

            if not dataframe.isnull().values.any():
                print("No missing values!")  # pragma: no cover
            else:
                total = dataframe.isnull().sum().sort_values(ascending=False)
                percent = (
                    dataframe.isnull().sum() / dataframe.isnull().count()
                ).sort_values(ascending=False)
                missing_data = pd.concat(
                    [total, percent], axis=1, keys=["Total", "Percent"]
                )

                missing_df.append(
                    missing_data.style.format({"Percent": "{:.2%}"}).set_caption(
                        caption
                    )
                )

        def multi_table(table_list):
            """ Acceps a list of IpyTable objects and returns a table which contains each IpyTable in a cell
            """
            return HTML(
                '<table><tr style="background-color:white;">'
                + "".join(
                    [
                        "<td style='padding-right:25px'>"
                        + table._repr_html_()
                        + "</td>"
                        for table in table_list
                    ]
                )
                + "</tr></table>"
            )

        if shell == "ZMQInteractiveShell":
            display(multi_table(missing_df))
        else:
            for table in missing_df:
                print(table)

    def copy(self):
        """
        Returns deep copy of object.
        
        Returns
        -------
        Object
            Deep copy of object
        """

        return copy.deepcopy(self)

    def log(self, msg: str):
        """
        Logs notes to the report file
        """

        self.report.log(msg)

    def standardize_column_names(self):
        """
        Utility function that standardizes all column names to lowercase and underscores for spaces.

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.standardize_column_names()
        """

        new_column_names = {}
        pattern = re.compile("\W+")

        for name in self.x_train.columns:
            new_column_names[name] = re.sub(pattern, "_", name.lower())

        if self.target_field is not None:
            self.target_field = re.sub(pattern, "_", self.target_field.lower())

        self.col_mapping = new_column_names

        self.x_train.rename(columns=new_column_names, inplace=True)

        if self.x_test is not None:
            self.x_test.rename(columns=new_column_names, inplace=True)

        return self.copy()

    def expand_json_column(self, col):
        """
        Utility function that expands a column that has JSON elements into columns, where each JSON key is a column. 

        Parameters
        ----------
        cols: str
            Column in the data that has the nested data.

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.expand_json_column('col1')
        """

        from pandas.io.json import json_normalize

        df = json_normalize(self.x_train[col], sep="_")
        self.x_train.drop(col, axis=1, inplace=True)
        self.x_train = pd.concat([self.x_train, df], axis=1)

        if self.x_test is not None:
            df = json_normalize(self.x_test[col], sep="_")
            self.x_test.drop(col, axis=1, inplace=True)
            self.x_test = pd.concat([self.x_test, df], axis=1)

        return self.copy()

    def search(self, *values, not_equal=False, replace=False):
        """
        Searches the entire dataset for specified value(s) and returns rows that contain the values.
        
        Parameters
        ----------
        values : Any
            Value to search for in dataframe

        not_equal : bool, optional
            True if you want filter by values in the dataframe that are not equal to the value provided, by default False

        replace : bool, optional
            Whether to permanently transform your data, by default False

        Returns
        -------
        Data or DataFrame:
            Returns a deep copy of the Data object or a DataFrame if replace is False.

        Examples
        --------
        >>> data.search('aethos')
        """

        # TODO: Refactor this to take in boolean expressions

        if not values:
            return ValueError("Please provide value to search.")

        if replace:
            if not_equal:
                self.x_train = self.x_train[self.x_train.isin(list(values)).any(axis=1)]

                if self.x_test is not None:
                    self.x_test = self.x_test[
                        self.x_test.isin(list(values)).any(axis=1)
                    ]
            else:
                self.x_train = self.x_train[self.x_train.isin(list(values)).any(axis=1)]

                if self.x_test is not None:
                    self.x_test = self.x_test[
                        self.x_test.isin(list(values)).any(axis=1)
                    ]

            return self.copy()
        else:
            data = self.x_train.copy()

            if not not_equal:
                data = data[data.isin(list(values))].dropna(how="all")
            else:
                data = data[~data.isin(list(values))].dropna(how="all")

            return data

    def where(self, *filter_columns, **columns):
        """
        Filters the dataframe down for highlevel analysis. 

        Can only handle '==', for more complex queries, interact with pandas.
        
        Parameters
        ----------
        filter_columns : str(s)
            Columns you want to see at the end result

        columns : key word arguments
            Columns and the associated value to filter on.
            Columns can equal a value or a list of values to include.
        
        Returns
        -------
        Dataframe
            A view of your data or training data

        Examples
        --------
        '''This translates to your data where col2 is equal to 3 and col 3 is equal to 4 and column 4 is equal to 1, 2 or 3.
        The col1 specifies that this this is the only column you want to see at the output.'''
        >>> data.where('col1', col2=3, col3=4, col4=[1,2,3])
        """

        filtered_data = self.x_train.copy()

        for col in columns.keys():
            if isinstance(columns[col], list):
                filtered_data = filtered_data[filtered_data[col].isin(columns[col])]
            else:
                filtered_data = filtered_data[filtered_data[col] == columns[col]]

        if filter_columns:
            return filtered_data[list(filter_columns)]
        else:
            return filtered_data

    def groupby(self, *groupby, replace=False):
        """
        Groups data by the provided columns.
        
        Parameters
        ----------
        groupby : str(s)
            Columns to group the data by.

        replace : bool, optional
            Whether to permanently transform your data, by default False
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.groupby('col1', 'col2')
        """

        if not groupby:
            return ValueError("Please provided columns to groupby.")

        if replace:
            self.x_train = self.x_train.groupby(list(groupby))

            if self.x_test is not None:
                self.x_test = self.x_test.groupby(list(groupby))

            return self.copy()
        else:
            data = self.x_train.copy()

            return data.groupby(list(groupby))

    def groupby_analysis(self, groupby: list, *cols, data_filter=None):
        """
        TODO: Refactor this to a class method.

        Groups your data and then provides descriptive statistics for the other columns on the grouped data.

        For numeric data, the descriptive statistics are:

            - count
            - min
            - max
            - mean
            - standard deviation
            - variance
            - median
            - most common
            - sum
            - Median absolute deviation
            - number of unique values

        For other types of data:

            - count
            - most common
            - number of unique values
        
        Parameters
        ----------
        groupby : list
            List of columns to groupby.

        cols : str(s)
            Columns you want statistics on, if none are provided, it will provide statistics for every column.

        data_filter : Dataframe, optional
            Filtered dataframe, by default None
        
        Returns
        -------
        Dataframe
            Dataframe of grouped columns and statistics for each column.

        Examples
        --------
        >>> data.groupby_analysis()
        >>> data.groupby_analysis('col1', 'col2')
        """

        analysis = {}
        numeric_analysis = [
            "count",
            "min",
            "max",
            "mean",
            "std",
            "var",
            "median",
            ("most_common", lambda x: pd.Series.mode(x)[0]),
            "sum",
            "mad",
            "nunique",
        ]
        other_analysis = [
            "count",
            ("most_common", lambda x: pd.Series.mode(x)[0]),
            "nunique",
        ]

        list_of_cols = _get_columns(list(cols), self.x_train)

        if isinstance(data_filter, pd.DataFrame):
            data = data_filter
        else:
            data = self.x_train.copy()

        for col in list_of_cols:
            if col not in groupby:
                # biufc - bool, int, unsigned, float, complex
                if data[col].dtype.kind in "biufc":
                    analysis[col] = numeric_analysis
                else:
                    analysis[col] = other_analysis

        analyzed_data = data.groupby(groupby).agg(analysis)

        return analyzed_data

    def data_report(self, title="Profile Report", output_file="", suppress=False):
        """
        Generates a full Exploratory Data Analysis report using Pandas Profiling.

        Credits: https://github.com/pandas-profiling/pandas-profiling
        
        For each column the following statistics - if relevant for the column type - are presented in an interactive HTML report:

        - Essentials: type, unique values, missing values
        - Quantile statistics like minimum value, Q1, median, Q3, maximum, range, interquartile range
        - Descriptive statistics like mean, mode, standard deviation, sum, median absolute deviation, coefficient of variation, kurtosis, skewness
        - Most frequent values
        - Histogram
        - Correlations highlighting of highly correlated variables, Spearman, Pearson and Kendall matrices
        - Missing values matrix, count, heatmap and dendrogram of missing values
        
        Parameters
        ----------
        title : str, optional
            Title of the report, by default 'Profile Report'

        output_file : str, optional
            File name of the output file for the report, by default ''

        suppress : bool, optional
            True if you do not want to display the report, by default False
        
        Returns
        -------
        HTML display of Exploratory Data Analysis report

        Examples
        --------
        >>> data.data_report()
        >>> data.data_report(title='Titanic EDA', output_file='titanic.html')
        """

        import pandas_profiling

        if shell == "ZMQInteractiveShell":  # pragma : no cover
            report = self.x_train.profile_report(
                title=title, style={"full_width": True}
            )
        else:
            report = self.x_train.profile_report(title=title)

        if output_file:
            report.to_file(output_file=output_file)

        if not suppress:
            return report

    def describe(self, dataset="train"):
        """
        Describes your dataset using the DataFrameSummary library with basic descriptive info.
        Extends the DataFrame.describe() method to give more info.

        Credits go to @mouradmourafiq for his pandas-summary library.
        
        Parameters
        ----------
        dataset : str, optional
            Type of dataset to describe. Can either be `train` or `test`.
            If you are using the full dataset it will automatically describe
            your full dataset no matter the input, 
            by default 'train'
        
        Returns
        -------
        DataFrame
            Dataframe describing your dataset with basic descriptive info

        Examples
        ---------
        >>> data.describe()
        """

        from pandas_summary import DataFrameSummary

        if dataset == "train":
            x_train_summary = DataFrameSummary(self.x_train)

            return x_train_summary.summary()
        else:
            x_test_summary = DataFrameSummary(self.x_test)

            return x_test_summary.summary()

    def column_info(self, dataset="train"):
        """
        Describes your columns using the DataFrameSummary library with basic descriptive info.

        Credits go to @mouradmourafiq for his pandas-summary library.

        Info
        ----
        counts
        uniques
        missing
        missing_perc
        types
        
        Parameters
        ----------
        dataset : str, optional
            Type of dataset to describe. Can either be `train` or `test`.
            If you are using the full dataset it will automatically describe
            your full dataset no matter the input, 
            by default 'train'
        
        Returns
        -------
        DataFrame
            Dataframe describing your columns with basic descriptive info

        Examples
        ---------
        >>> data.column_info()
        """

        from pandas_summary import DataFrameSummary

        if dataset == "train":
            x_train_summary = DataFrameSummary(self.x_train)

            return x_train_summary.columns_stats
        else:
            x_test_summary = DataFrameSummary(self.x_test)

            return x_test_summary.columns_stats

    def describe_column(self, column, dataset="train"):
        """
        Analyzes a column and reports descriptive statistics about the columns.

        Credits go to @mouradmourafiq for his pandas-summary library.

        Statistics
        ----------
        std                                      
        max                                      
        min                                      
        variance                                 
        mean
        mode                                     
        5%                                       
        25%                                      
        50%                                      
        75%                                      
        95%                                      
        iqr                                      
        kurtosis                                 
        skewness                                 
        sum                                      
        mad                                      
        cv                                       
        zeros_num                                
        zeros_perc                               
        deviating_of_mean                        
        deviating_of_mean_perc                   
        deviating_of_median                      
        deviating_of_median_perc                 
        top_correlations                         
        counts                                   
        uniques                                  
        missing                                  
        missing_perc                             
        types                            
        
        Parameters
        ----------
        column : str
            Column in your dataset you want to analze.

        dataset : str, optional
            Type of dataset to describe. Can either be `train` or `test`.
            If you are using the full dataset it will automatically describe
            your full dataset no matter the input, 
            by default 'train'
        
        Returns
        -------
        dict
            Dictionary mapping a statistic and its value for a specific column
            
        Examples
        --------
        >>> data.describe_column('col1')
        """

        from pandas_summary import DataFrameSummary

        if dataset == "train":
            x_train_summary = DataFrameSummary(self.x_train)

            return x_train_summary[column]
        else:
            x_test_summary = DataFrameSummary(self.x_test)

            return x_test_summary[column]

    def drop(self, *drop_columns, keep=[], regexp="", reason=""):
        """
        Drops columns from the dataframe.
        
        Parameters
        ----------
        keep : list: optional
            List of columns to not drop, by default []

        regexp : str, optional
            Regular Expression of columns to drop, by default ''

        reason : str, optional
            Reasoning for dropping columns, by default ''

        Column names must be provided as strings that exist in the data.
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.drop('A', 'B', reason="Columns were unimportant")
        >>> data.drop('col1', keep=['col2'], regexp=r"col*") # Drop all columns that start with "col" except column 2
        >>> data.drop(keep=['A']) # Drop all columns except column 'A'
        >>> data.drop(regexp=r'col*') # Drop all columns that start with 'col'       
        """

        if not isinstance(keep, list):
            raise TypeError("Keep parameter must be a list.")

        # Handles if columns do not exist in the dataframe
        data_columns = self.x_train.columns
        regex_columns = []

        if regexp:
            regex = re.compile(regexp)
            regex_columns = list(filter(regex.search, data_columns))

        drop_columns = set(drop_columns).union(regex_columns)

        # If there are columns to be dropped, exclude the ones in the keep list
        # If there are no columns to be dropped, drop everything except the keep list
        if drop_columns:
            drop_columns = list(drop_columns.difference(keep))
        else:
            keep = set(data_columns).difference(keep)
            drop_columns = list(drop_columns.union(keep))

        self.x_train = self.x_train.drop(drop_columns, axis=1)

        if self.x_test is not None:
            self.x_test = self.x_test.drop(drop_columns, axis=1)

        if self.report is not None:
            self.report.log(f'Dropped columns: {", ".join(drop_columns)}. {reason}')

        return self.copy()

    def encode_target(self):
        """
        Encodes target variables with value between 0 and n_classes-1.

        Running this function will automatically set the corresponding mapping for the target variable mapping number to the original value.

        Note that this will not work if your test data will have labels that your train data does not.        

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.encode_target()
        """

        if not self.target_field:
            raise ValueError(
                "Please set the `target_field` field variable before encoding."
            )

        (self.x_train, self.x_test, self.target_mapping,) = label_encoder(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=[self.target_field],
            target=True,
        )

        for k, v in self.target_mapping.items():
            print(f"{k}: {v}")

        if self.report is not None:
            self.report.log("Encoded the target variable as numeric values.")

        return self.copy()

    def to_csv(self, name: str, index=False, **kwargs):
        """
        Write data to csv with the name and path provided.

        The function will automatically add '.csv' to the end of the name.

        By default it writes 10000 rows at a time to file to consider memory on different machines.

        Training data will end in '_train.csv' andt test data will end in '_test.csv'.

        For a full list of keyword args for writing to csv please see the following link: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_csv.html
        
        Parameters
        ----------
        name : str
            File path

        index : bool, optional
            True to write 'index' column, by default False

        Examples
        --------
        >>> data.to_csv('titanic')
        """

        index = kwargs.pop("index", index)
        chunksize = kwargs.pop("chunksize", 10000)

        self.x_train.to_csv(
            name + "_train.csv", index=index, chunksize=chunksize, **kwargs
        )

        if self.x_test is not None:
            self.x_test.to_csv(
                name + "_test.csv", index=index, chunksize=chunksize, **kwargs
            )

    def checklist(self):
        """
        Displays a checklist dashboard with reminders for a Data Science project.

        Examples
        --------
        >>> data.checklist()
        """

        data_checkboxes = []
        clean_checkboxes = []
        analysis_checkboxes = [
            [widgets.Label(value="Univariate Analysis")],
            [widgets.Label(value="Multivariate Analysis")],
            [widgets.Label(value="Timeseries Analysis")],
        ]
        issue_checkboxes = []
        preparation_checkboxes = []

        for item in DATA_CHECKLIST:
            data_checkboxes.append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )
        data_box = widgets.VBox(data_checkboxes)

        for item in CLEANING_CHECKLIST:
            clean_checkboxes.append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )
        clean_box = widgets.VBox(clean_checkboxes)

        for item in UNI_ANALYSIS_CHECKLIST:
            analysis_checkboxes[0].append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )
        uni_box = widgets.VBox(analysis_checkboxes[0])

        for item in MULTI_ANALYSIS_CHECKLIST:
            analysis_checkboxes[1].append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )

        multi_box = widgets.VBox(analysis_checkboxes[1])

        analysis_box = widgets.HBox([uni_box, multi_box])

        for item in ISSUES_CHECKLIST:
            issue_checkboxes.append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )
        issue_box = widgets.VBox(issue_checkboxes)

        for item in PREPARATION_CHECKLIST:
            preparation_checkboxes.append(
                widgets.Checkbox(description=item, layout=Layout(width="100%"))
            )
        prep_box = widgets.VBox(preparation_checkboxes)

        tab_list = [data_box, clean_box, analysis_box, issue_box, prep_box]

        tab = widgets.Tab()
        tab.children = tab_list
        tab.set_title(0, "Data")
        tab.set_title(1, "Cleaning")
        tab.set_title(2, "Analysis")
        tab.set_title(3, "Issues")
        tab.set_title(4, "Preparation")

        display(tab)

    def to_df(self):
        """
        Return Dataframes for x_train and x_test if it exists.

        Returns
        -------
        Dataframe, *Dataframe
            Transformed dataframe with rows with a missing values in a specific column are missing

            Returns 2 Dataframes test if x_test is provided.

        Examples
        --------
        >>> data.to_df()
        """

        if self.x_test is None:
            return self.x_train
        else:
            return self.x_train, self.x_test



"""
Implemented pandas version of package wide configuration: https://github.com/pandas-dev/pandas/blob/f9f95c0762af2c114ed27d21288c9d25ab49634d/pandas/_config/config.py

Overview
========
This module supports the following requirements:
- options are referenced using keys in dot.notation, e.g. "x.y.option - z".
- keys are case-insensitive.
- functions should accept partial/regex keys, when unambiguous.
- options can be registered by modules at import time.
- options can be registered at init-time (via core.config_init)
- options have a default value, and (optionally) a description and
  validation function associated with them.
- options can be deprecated, in which case referencing them
  should produce a warning.
- deprecated options can optionally be rerouted to a replacement
  so that accessing a deprecated option reroutes to a differently
  named option.
- options can be reset to their default value.
- all option can be reset to their default value at once.
- you can register a callback to be invoked when the option value
  is set or reset. Changing the stored value is considered misuse, but
  is not verboten.

Implementation
==============

- "Registered options" and "Deprecated options" have metadata associated
  with them, which are stored in auxiliary dictionaries keyed on the
  fully-qualified key, e.g. "x.y.z.option".

- the config_init module is imported by the package's __init__.py file.
  placing any register_option() calls there will ensure those options
  are available as soon as pandas is loaded. If you use register_option
  in a module, it will only be available after that module is imported,
  which you should be aware of.

- `config_prefix` is a context_manager (for use with the `with` keyword)
  which can save developers some typing, see the docstring.
"""

import re
from collections import namedtuple
from contextlib import contextmanager

_registered_options = {}
_global_config = {}

RegisteredOption = namedtuple("RegisteredOption", "key default doc validator cb")


class OptionError(AttributeError, KeyError):
    """Exception for pandas.options, backwards compatible with KeyError
    checks
    """


def register_option(key, default, doc="", validator=None, cb=None):
    """Register an option in the package-wide pandas config object
    Parameters
    ----------
    key       - a fully-qualified key, e.g. "x.y.option - z".
    default    - the default value of the option
    doc       - a string description of the option
    validator - a function of a single argument, should raise `ValueError` if
                called with a value which is not a legal value for the option.
    cb        - a function of a single argument "key", which is called
                immediately after an option value is set/reset. key is
                the full name of the option.
    Returns
    -------
    Nothing.
    
    Raises
    ------
    ValueError if `validator` is specified and `default` is not a valid value.
    """
    import tokenize
    import keyword

    key = key.lower()

    if key in _registered_options:
        msg = "Option '{key}' has already been registered"
        raise OptionError(msg.format(key=key))

    # the default value should be legal
    if validator:
        validator(default)

    _global_config[key] = default  # initialize

    # save the option metadata
    _registered_options[key] = RegisteredOption(
        key=key, default=default, doc=doc, validator=validator, cb=cb
    )


def _select_options(pat):
    """returns a list of keys matching `pat`
    if pat=="all", returns all registered options
    """

    # short-circuit for exact key
    if pat in _registered_options:
        return [pat]

    # else look through all of them
    keys = sorted(_registered_options.keys())
    if pat == "all":  # reserved key
        return keys

    return [k for k in keys if re.search(pat, k, re.I)]


def _get_single_key(pat):
    keys = _select_options(pat)
    if len(keys) == 0:
        raise OptionError("No such keys(s): {pat!r}".format(pat=pat))
    if len(keys) > 1:
        raise OptionError("Pattern matched multiple keys")
    key = keys[0]

    return key


def _get_registered_option(key):
    """
    Retrieves the option metadata if `key` is a registered option.

    Returns
    -------
    RegisteredOption (namedtuple) if key is deprecated, None otherwise
    """

    return _registered_options.get(key)


def _get_option(pat):
    key = _get_single_key(pat)

    return _global_config[key]


def _set_option(*args, **kwargs):
    # must at least 1 arg deal with constraints later
    nargs = len(args)
    if not nargs or nargs % 2 != 0:
        raise ValueError("Must provide an even number of non-keyword arguments")

    if kwargs:
        msg = '_set_option() got an unexpected keyword argument "{kwarg}"'
        raise TypeError(msg.format(list(kwargs.keys())[0]))

    for k, v in zip(args[::2], args[1::2]):
        key = _get_single_key(k)

        o = _get_registered_option(key)
        if o and o.validator:
            o.validator(v)

        _global_config[key] = v

        if o.cb:
            o.cb(key)


def _build_option_description(k):
    """ Builds a formatted description of a registered option and prints it """

    o = _get_registered_option(k)

    s = "{k} ".format(k=k)

    if o.doc:
        s += "\n".join(o.doc.strip().split("\n"))
    else:
        s += "No description available."

    if o:
        s += "\n    [default: {default}] [currently: {current}]".format(
            default=o.default, current=_get_option(k)
        )

    return s


def _describe_option(pat="", _print_desc=True):

    keys = _select_options(pat)
    if len(keys) == 0:
        raise OptionError("No such keys(s)")

    s = ""
    for k in keys:  # filter by pat
        s += _build_option_description(k)

    if _print_desc:
        print(s)
    else:
        return s


def _reset_option(pat):

    keys = _select_options(pat)

    if len(keys) == 0:
        raise OptionError("No such keys(s)")

    if len(keys) > 1 and len(pat) < 4 and pat != "all":
        raise ValueError(
            "You must specify at least 4 characters when "
            "resetting multiple keys, use the special keyword "
            '"all" to reset all the options to their default '
            "value"
        )

    for k in keys:
        _set_option(k, _registered_options[k].default)


def pp_options_list(keys, width=80, _print=False):
    """ Builds a concise listing of available options, grouped by prefix """

    from textwrap import wrap
    from itertools import groupby

    def pp(name, ks):
        pfx = "- " + name + ".[" if name else ""
        ls = wrap(
            ", ".join(ks),
            width,
            initial_indent=pfx,
            subsequent_indent="  ",
            break_long_words=False,
        )
        if ls and ls[-1] and name:
            ls[-1] = ls[-1] + "]"
        return ls

    ls = []
    singles = [x for x in sorted(keys) if x.find(".") < 0]
    if singles:
        ls += pp("", singles)
    keys = [x for x in keys if x.find(".") >= 0]

    for k, g in groupby(sorted(keys), lambda x: x[: x.rfind(".")]):
        ks = [x[len(k) + 1 :] for x in list(g)]
        ls += pp(k, ks)
    s = "\n".join(ls)
    if _print:
        print(s)
    else:
        return s


class DictWrapper:
    """ provide attribute-style access to a nested dict"""

    def __init__(self, d, prefix=""):
        object.__setattr__(self, "d", d)
        object.__setattr__(self, "prefix", prefix)

    def __setattr__(self, key, val):
        prefix = object.__getattribute__(self, "prefix")
        if prefix:
            prefix += "."
        prefix += key
        # you can't set new keys
        # can you can't overwrite subtrees
        if key in self.d and not isinstance(self.d[key], dict):
            _set_option(prefix, val)
        else:
            raise OptionError("You can only set the value of existing options")

    def __getattr__(self, key):
        prefix = object.__getattribute__(self, "prefix")
        if prefix:
            prefix += "."
        prefix += key
        try:
            v = object.__getattribute__(self, "d")[key]
        except KeyError:
            raise OptionError("No such option")
        if isinstance(v, dict):
            return DictWrapper(v, prefix)
        else:
            return _get_option(prefix)

    def __dir__(self):
        return list(self.d.keys())


class CallableDynamicDoc:
    def __init__(self, func, doc_templ):
        self.__doc_tmpl__ = doc_templ
        self.__func__ = func

    def __call__(self, *args, **kwargs):
        return self.__func__(*args, **kwargs)

    @property
    def __doc__(self):
        opts_desc = _describe_option("all", _print_desc=False)
        opts_list = pp_options_list(list(_registered_options.keys()))

        return self.__doc_tmpl__.format(opts_desc=opts_desc, opts_list=opts_list)


_get_option_tmpl = """
get_option(pat)

Retrieves the value of the specified option.

Available options:

{opts_list}

Parameters
----------
pat : str
    Regexp which should match a single option.
    Note: partial matches are supported for convenience, but unless you use the
    full option name (e.g. x.y.z.option_name), your code may break in future
    versions if new options with similar names are introduced.

Returns
-------
result : the value of the option

Raises
------
OptionError : if no such option exists

Notes
-----
The available options with its descriptions:

{opts_desc}
"""

_set_option_tmpl = """
set_option(pat, value)

Sets the value of the specified option.

Available options:

{opts_list}

Parameters
----------
pat : str
    Regexp which should match a single option.
    Note: partial matches are supported for convenience, but unless you use the
    full option name (e.g. x.y.z.option_name), your code may break in future
    versions if new options with similar names are introduced.
value : object
    New value of option.

Returns
-------
None

Raises
------
OptionError if no such option exists

Notes
-----
The available options with its descriptions:

{opts_desc}
"""

_describe_option_tmpl = """
describe_option(pat, _print_desc=False)

Prints the description for one or more registered options.

Call with not arguments to get a listing for all registered options.

Available options:

{opts_list}

Parameters
----------
pat : str
    Regexp pattern. All matching keys will have their description displayed.
_print_desc : bool, default True
    If True (default) the description(s) will be printed to stdout.
    Otherwise, the description(s) will be returned as a unicode string
    (for testing).

Returns
-------
None by default, the description(s) as a unicode string if _print_desc
is False

Notes
-----
The available options with its descriptions:

{opts_desc}
"""

_reset_option_tmpl = """
reset_option(pat)

Reset one or more options to their default value.

Pass "all" as argument to reset all options.

Available options:

{opts_list}

Parameters
----------
pat : str/regex
    If specified only options matching `prefix*` will be reset.
    Note: partial matches are supported for convenience, but unless you
    use the full option name (e.g. x.y.z.option_name), your code may break
    in future versions if new options with similar names are introduced.

Returns
-------
None

Notes
-----
The available options with its descriptions:

{opts_desc}
"""

get_option = CallableDynamicDoc(_get_option, _get_option_tmpl)
set_option = CallableDynamicDoc(_set_option, _set_option_tmpl)
reset_option = CallableDynamicDoc(_reset_option, _reset_option_tmpl)
describe_option = CallableDynamicDoc(_describe_option, _describe_option_tmpl)
options = DictWrapper(_global_config)


@contextmanager
def config_prefix(prefix):
    """contextmanager for multiple invocations of API with a common prefix
    supported API functions: (register / get / set )__option
    Warning: This is not thread - safe, and won't work properly if you import
    the API functions into your module using the "from x import y" construct.
    Example:
    import pandas._config.config as cf
    with cf.config_prefix("display.font"):
        cf.register_option("color", "red")
        cf.register_option("size", " 5 pt")
        cf.set_option(size, " 6 pt")
        cf.get_option(size)
        ...
        etc'
    will register options "display.font.color", "display.font.size", set the
    value of "display.font.size"... and so on.
    """

    # Note: reset_option relies on set_option, and on key directly
    # it does not fit in to this monkey-patching scheme

    global register_option, get_option, set_option, reset_option

    def wrap(func):
        def inner(key, *args, **kwds):
            pkey = "{prefix}.{key}".format(prefix=prefix, key=key)
            return func(pkey, *args, **kwds)

        return inner

    _register_option = register_option
    _get_option = get_option
    _set_option = set_option
    set_option = wrap(set_option)
    get_option = wrap(get_option)
    register_option = wrap(register_option)
    yield None
    set_option = _set_option
    get_option = _get_option
    register_option = _register_option


def is_type_factory(_type):
    """
    Parameters
    ----------
    `_type` - a type to be compared against (e.g. type(x) == `_type`)
    Returns
    -------
    validator - a function of a single argument x , which raises
                ValueError if type(x) is not equal to `_type`
    """

    def inner(x):
        if type(x) != _type:
            msg = "Value must have type '{typ!s}'"
            raise ValueError(msg.format(typ=_type))

    return inner


is_bool = is_type_factory(bool)
is_list = is_type_factory(list)



import os
import unittest

import numpy as np
import pandas as pd

from aethos import Data
from aethos.reporting.report import Report


class TestReport(unittest.TestCase):
    def test_report_writing(self):
        header = "Test"
        contents = "This is a test."

        report = Report("test")
        report.write_header(header)
        report.write_contents(contents)

        with open(report.filename) as f:
            content = f.read()

        os.remove(report.filename)

        self.assertTrue(True)

    def test_report_cleaning_technique(self):

        int_missing_data = np.array([(1, 0, 0), (0, None, None), (None, None, None)])
        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(int_missing_data, columns=columns)

        clean = Data(
            x_train=data, test_split_percentage=0.5, split=False, report_name="test"
        )
        clean.drop_column_missing_threshold(0.5)

        with open(clean.report.filename) as f:
            content = f.read()
        validate = "col2" in content and "col3" in content

        os.remove(clean.report.filename)

        self.assertTrue(validate)

    def test_report_cleaning_new_category(self):

        missing_data = [[1.0, "Green", 2], [1.0, "Other", 1], [np.nan, np.nan, np.nan]]

        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(missing_data, columns=columns)

        clean = Data(
            x_train=data, test_split_percentage=0.5, split=False, report_name="test"
        )
        clean_data = clean.replace_missing_new_category()

        with open(clean.report.filename) as f:
            content = f.read()
        validate = "col1" in content and "col2" in content and "col3" in content

        os.remove(clean.report.filename)

        self.assertTrue(validate)

    def test_report_preprocessing_standardize(self):

        unnormal_data = [[5.0, 3, 1], [2.0, 2, 1], [10.0, 1, 1]]

        columns = ["col1", "col2", "col3"]
        data = pd.DataFrame(unnormal_data, columns=columns)

        preprocess = Data(
            x_train=data, test_split_percentage=0.5, split=False, report_name="test"
        )
        preprocess.normalize_numeric()

        with open(preprocess.report.filename) as f:
            content = f.read()
        validate = "col1" in content and "col2" in content and "col3" in content

        os.remove(preprocess.report.filename)

        self.assertTrue(validate)

    def test_report_feature_bow(self):

        list_of_sentences = ["Hi my name is pyml", "Hi name pyml"]

        columns = ["text"]
        data = pd.DataFrame(list_of_sentences, columns=columns)

        feature = Data(
            x_train=data, test_split_percentage=0.5, split=False, report_name="test"
        )
        feature.bag_of_words()

        with open(feature.report.filename) as f:
            content = f.read()
        validate = "representation" in content

        os.remove(feature.report.filename)

        self.assertTrue(validate)


if __name__ == "__main__":
    unittest.main()



from aethos.config import technique_reason_repo
from aethos.preprocessing import categorical as cat
from aethos.preprocessing import numeric as num
from aethos.preprocessing import text
from aethos.util import (
    _input_columns,
    _numeric_input_conditions,
)


class Preprocess(object):
    def normalize_numeric(self, *list_args, list_of_cols=[], **normalize_params):
        """
        Function that normalizes all numeric values between 2 values to bring features into same domain.
        
        If `list_of_cols` is not provided, the strategy will be applied to all numeric columns.

        If a list of columns is provided use the list, otherwise use arguments.

        For more info please see: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html#sklearn.preprocessing.MinMaxScaler 

        This function can be found in `preprocess/numeric.py`     
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        feature_range : tuple(int or float, int or float), optional
            Min and max range to normalize values to, by default (0, 1)

        normalize_params : dict, optional
            Parmaters to pass into MinMaxScaler() constructor from Scikit-Learn
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.normalize_numeric('col1')
        >>> data.normalize_numeric(['col1', 'col2'])
        """

        report_info = technique_reason_repo["preprocess"]["numeric"]["standardize"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = num.scale(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            method="minmax",
            **normalize_params,
        )

        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def normalize_quantile_range(self, *list_args, list_of_cols=[], **robust_params):
        """
        Scale features using statistics that are robust to outliers.

        This Scaler removes the median and scales the data according to the quantile range (defaults to IQR: Interquartile Range).
        The IQR is the range between the 1st quartile (25th quantile) and the 3rd quartile (75th quantile).

        Standardization of a dataset is a common requirement for many machine learning estimators.
        Typically this is done by removing the mean and scaling to unit variance.
        However, outliers can often influence the sample mean / variance in a negative way.
        In such cases, the median and the interquartile range often give better results.
        
        If `list_of_cols` is not provided, the strategy will be applied to all numeric columns.

        If a list of columns is provided use the list, otherwise use arguments.

        For more info please see: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.RobustScaler.html#sklearn.preprocessing.RobustScaler

        This function can be found in `preprocess/numeric.py`     
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        with_centering : boolean, True by default
            If True, center the data before scaling.
            This will cause transform to raise an exception when attempted on sparse matrices,
            because centering them entails building a dense matrix which in common use cases is likely to be too large to fit in memory.
        
        with_scaling : boolean, True by default
            If True, scale the data to interquartile range.

        quantile_range : tuple (q_min, q_max), 0.0 < q_min < q_max < 100.0
            Default: (25.0, 75.0) = (1st quantile, 3rd quantile) = IQR Quantile range used to calculate scale_.

        robust_params : dict, optional
            Parmaters to pass into MinMaxScaler() constructor from Scikit-Learn
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.normalize_quantile_range('col1')
        >>> data.normalize_quantile_range(['col1', 'col2'])
        """

        report_info = technique_reason_repo["preprocess"]["numeric"]["robust"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = num.scale(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            method="robust",
            **robust_params,
        )

        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def normalize_log(self, *list_args, list_of_cols=[], base=1):
        """
        Scales data logarithmically.

        Options are 1 for natural log, 2 for base2, 10 for base10.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        base : str, optional
            Base to logarithmically scale by, by default ''
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.normalize_log('col1')
        >>> data.normalize_log(['col1', 'col2'], base=10)
        """

        report_info = technique_reason_repo["preprocess"]["numeric"]["log"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = num.log_scale(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            base=base,
        )

        if self.report is not None:
            if list_of_cols:
                self.report.report_technique(report_info, list_of_cols)
            else:
                list_of_cols = _numeric_input_conditions(list_of_cols, self.x_train)
                self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def split_sentences(self, *list_args, list_of_cols=[], new_col_name="_sentences"):
        """
        Splits text data into sentences and saves it into another column for analysis.

        If a list of columns is provided use the list, otherwise use arguments.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_sentences`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.split_sentences('col1')
        >>> data.split_sentences(['col1', 'col2'])
        """

        report_info = technique_reason_repo["preprocess"]["text"]["split_sentence"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = text.split_sentences(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def stem_nltk(
        self, *list_args, list_of_cols=[], stemmer="porter", new_col_name="_stemmed"
    ):
        """
        Transforms text to their word stem, base or root form. 
        For example:
            dogs --> dog
            churches --> church
            abaci --> abacus
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        stemmer : str, optional
            Type of NLTK stemmer to use, by default porter

            Current stemming implementations:
                - porter
                - snowball

            For more information please refer to the NLTK stemming api https://www.nltk.org/api/nltk.stem.html

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_stemmed`
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.stem_nltk('col1')
        >>> data.stem_nltk(['col1', 'col2'], stemmer='snowball')
        """

        report_info = technique_reason_repo["preprocess"]["text"]["stem"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        self.x_train, self.x_test = text.nltk_stem(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            stemmer=stemmer,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def split_words_nltk(
        self, *list_args, list_of_cols=[], regexp="", new_col_name="_tokenized"
    ):
        """
        Splits text into its words using nltk punkt tokenizer by default. 
    
        Default is by spaces and punctuation but if a regex expression is provided, it will use that.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        regexp : str, optional
            Regex expression used to define what a word is.

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_tokenized`
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.split_words_nltk('col1')
        >>> data.split_words_nltk(['col1', 'col2'])
        """

        report_info = technique_reason_repo["preprocess"]["text"]["split_words"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.nltk_word_tokenizer(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            regexp=regexp,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def remove_stopwords_nltk(
        self, *list_args, list_of_cols=[], custom_stopwords=[], new_col_name="_rem_stop"
    ):
        """
        Removes stopwords following the nltk English stopwords list.

        A list of custom words can be provided as well, usually for domain specific words.

        Stop words are generally the most common words in a language
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        custom_stop_words : list, optional
            Custom list of words to also drop with the stop words, must be LOWERCASE, by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_rem_stop`
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.remove_stopwords_nltk('col1')
        >>> data.remove_stopwords_nltk(['col1', 'col2'])
        """

        report_info = technique_reason_repo["preprocess"]["text"]["remove_stopwords"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.nltk_remove_stopwords(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            custom_stopwords=custom_stopwords,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def remove_punctuation(
        self,
        *list_args,
        list_of_cols=[],
        regexp="",
        exceptions=[],
        new_col_name="_rem_punct",
    ):
        """
        Removes punctuation from every string entry.

        Defaults to removing all punctuation, but if regex of punctuation is provided, it will remove them.
        
        An example regex would be:

        (\w+\.|\w+)[^,] - Include all words and words with periods after them but don't include commas.
        (\w+\.)|(\w+), would also achieve the same result

        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        regexp : str, optional
            Regex expression used to define what to include.
            
        exceptions : list, optional
            List of punctuation to include in the text, by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_rem_punct`
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.remove_punctuation('col1')
        >>> data.remove_punctuation(['col1', 'col2'])
        >>> data.remove_punctuation('col1', regexp=r'(\w+\.)|(\w+)') # Include all words and words with periods after.
        """

        report_info = technique_reason_repo["preprocess"]["text"]["remove_punctuation"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.remove_punctuation(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            regexp=regexp,
            exceptions=exceptions,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def remove_numbers(self, *list_args, list_of_cols=[], new_col_name="_rem_num"):
        """
        Removes numbers from text in a column.
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_rem_num`

        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.remove_numbers('col1', new_col_name="text_wo_num)
        """

        report_info = technique_reason_repo["preprocess"]["text"]["remove_numbers"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        (self.x_train, self.x_test,) = text.remove_numbers(
            x_train=self.x_train,
            x_test=self.x_test,
            list_of_cols=list_of_cols,
            new_col_name=new_col_name,
        )

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()

    def clean_text(
        self,
        *list_args,
        list_of_cols=[],
        lower=True,
        punctuation=True,
        stopwords=True,
        stemmer=True,
        numbers=True,
        new_col_name="_clean",
    ):
        """
        Function that takes text and does the following:

        - Casts it to lowercase
        - Removes punctuation
        - Removes stopwords
        - Stems the text
        - Removes any numerical text
        
        Parameters
        ----------
        list_args : str(s), optional
            Specific columns to apply this technique to.

        list_of_cols : list, optional
            A list of specific columns to apply this technique to., by default []
        
        lower : bool, optional
            True to cast all text to lowercase, by default True
        
        punctuation : bool, optional
            True to remove punctuation, by default True
        
        stopwords : bool, optional
            True to remove stop words, by default True
        
        stemmer : bool, optional
            True to stem the data, by default True

        numbers : bool, optional
            True to remove numerical data, by default True

        new_col_name : str, optional
            New column name to be created when applying this technique, by default `COLUMN_clean`            
        
        Returns
        -------
        Data:
            Returns a deep copy of the Data object.

        Examples
        --------
        >>> data.clean_text('col1')
        >>> data.clean_text(['col1', 'col2'], lower=False)
        >>> data.clean_text(lower=False, stopwords=False, stemmer=False)
        """

        report_info = technique_reason_repo["preprocess"]["text"]["clean_text"]

        list_of_cols = _input_columns(list_args, list_of_cols)

        for col in list_of_cols:
            if new_col_name.startswith("_"):
                new_col_name = col + new_col_name

            self.x_train[new_col_name] = [
                text.process_text(
                    txt,
                    lower=lower,
                    punctuation=punctuation,
                    stopwords=stopwords,
                    stemmer=stemmer,
                    numbers=numbers,
                )
                for txt in self.x_train[col]
            ]

            if self.x_test is not None:
                self.x_test[new_col_name] = [
                    text.process_text(
                        txt,
                        lower=lower,
                        punctuation=punctuation,
                        stopwords=stopwords,
                        stemmer=stemmer,
                        numbers=numbers,
                    )
                    for txt in self.x_test[col]
                ]

        if self.report is not None:
            self.report.report_technique(report_info, list_of_cols)

        return self.copy()



import os
import sys
from subprocess import check_call

from setuptools import find_packages, setup
from setuptools.command.install import install

VERSION = "1.2.5"

pkgs = [
    "numpy==1.17.4",
    "pandas==0.25.1",
    "scikit-learn>=0.22",
    "textblob==0.15.3",
    "matplotlib<3.2.0",
    "pandas_summary",
    "ptitprince==0.1.5",
    "nltk==3.4.5",
    "ipython==7.10.1",
    "gensim==3.8.1",
    "pandas-profiling==2.3.0",
    "cookiecutter==1.6.0",
    "shap==0.33.0",
    "interpret==0.1.20",
    "yellowbrick==1.0.1",
    "spacy==2.2.3",
    "xgboost==0.90",
    "ipywidgets==7.5.1",
    "qgrid==1.1.1",
    "python-dateutil<2.8.1",
    "itables==0.2.1",
    "selenium==3.141.0",
    "python-docx==0.8.10",
    "graphviz==0.13.2",
    "pyldavis==2.1.2",
    "swifter==0.297",
    "lightgbm==2.3.1",
    "catboost==0.20.1",
    "mlflow==1.5.0",
    "statsmodels==0.10.2",
]

extras = {"ptmodels": ["transformers==2.3.0", "tensorflow==2.1.0"]}


class VerifyVersionCommand(install):
    """Custom command to verify that the git tag matches our version"""

    description = "verify that the git tag matches our version"

    def run(self):
        tag = os.getenv("CIRCLE_TAG")

        if tag != "v" + VERSION:
            info = "Git tag: {0} does not match the version of this app: {1}".format(
                tag, VERSION
            )
            sys.exit(info)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="aethos",
    url="https://github.com/Ashton-Sidhu/aethos",
    packages=find_packages(),
    author="Ashton Sidhu",
    author_email="ashton.sidhu1994@gmail.com",
    install_requires=pkgs,
    extras_require=extras,
    version=VERSION,
    license="GPL-3.0",
    description="A library of data science and machine learning techniques to help automate your workflow.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="datascience, machinelearning, automation, analysis",
    python_requires=">= 3.6",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Development Status :: 4 - Beta",
    ],
    cmdclass={"verify": VerifyVersionCommand},
    entry_points={"console_scripts": ["aethos=aethos.__main__:main"]},
)
